'''
Spatial analyses functions for Digital Earth Africa data.
'''

# Force GeoPandas to use Shapely instead of PyGEOS
# In a future release, GeoPandas will switch to using Shapely by default.
import os

os.environ['USE_PYGEOS'] = '0'

import multiprocessing as mp

import dask
import fiona
import geopandas as gpd
import numpy as np
import odc.geo.xr  # adds  `.odc.x` attributes to our xarray objects.
import pandas as pd
import rasterio.features
import scipy.interpolate
import xarray as xr
from datacube.api.query import query_group_by
from datacube.model.utils import xr_apply
from datacube.utils.cog import write_cog
from datacube.utils.geometry import CRS, Geometry
from geopy.geocoders import Nominatim
from rasterstats import zonal_stats
from shapely.geometry import LineString, MultiLineString, mapping, shape
from skimage.measure import find_contours, label


def add_geobox(ds, crs=None):
    """
    Ensure that an xarray DataArray has a GeoBox and .odc.* accessor
    using `odc.geo`.

    If `ds` is missing a Coordinate Reference System (CRS), this can be
    supplied using the `crs` param.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
        Input xarray object that needs to be checked for spatial
        information.
    crs : str, optional
        Coordinate Reference System (CRS) information for the input `ds`
        array. If `ds` already has a CRS, then `crs` is not required.
        Default is None.

    Returns
    -------
    xarray.Dataset or xarray.DataArray
        The input xarray object with added `.odc.x` attributes to access
        spatial information.

    """
    # If a CRS is not found, use custom provided CRS
    if ds.odc.crs is None and crs is not None:
        ds = ds.odc.assign_crs(crs)
    elif ds.odc.crs is None and crs is None:
        raise ValueError(
            "Unable to determine `ds`'s coordinate "
            "reference system (CRS). Please provide a "
            "CRS using the `crs` parameter "
            "(e.g. `crs='EPSG:3577'`)."
        )

    return ds


def xr_vectorize(
    da,
    attribute_col=None,
    crs=None,
    dtype="float32",
    output_path=None,
    verbose=True,
    **rasterio_kwargs,
):
    """
    Vectorises a raster ``xarray.DataArray`` into a vector
    ``geopandas.GeoDataFrame``.

    Parameters
    ----------
    da : xarray.DataArray
        The input ``xarray.DataArray`` data to vectorise.
    attribute_col : str, optional
        Name of the attribute column in the resulting
        ``geopandas.GeoDataFrame``. Values from ``da`` converted
        to polygons will be assigned to this column. If None,
        the column name will default to 'attribute'.
    crs : str or CRS object, optional
        If ``da``'s coordinate reference system (CRS) cannot be
        determined, provide a CRS using this parameter.
        (e.g. 'EPSG:3577').
    dtype : str, optional
         Data type  of  must be one of int16, int32, uint8, uint16,
         or float32
    output_path : string, optional
        Provide an optional string file path to export the vectorised
        data to file. Supports any vector file formats supported by
        ``geopandas.GeoDataFrame.to_file()``.
    verbose : bool, optional
        Print debugging messages. Default True.
    **rasterio_kwargs :
        A set of keyword arguments to ``rasterio.features.shapes``.
        Can include `mask` and `connectivity`.

    Returns
    -------
    gdf : geopandas.GeoDataFrame

    """

    # Add GeoBox and odc.* accessor to array using `odc-geo`
    da = add_geobox(da, crs)

    # Run the vectorizing function
    vectors = rasterio.features.shapes(
        source=da.data.astype(dtype), transform=da.odc.transform, **rasterio_kwargs
    )

    # Convert the generator into a list
    vectors = list(vectors)

    # Extract the polygon coordinates and values from the list
    polygons = [polygon for polygon, value in vectors]
    values = [value for polygon, value in vectors]

    # Convert polygon coordinates into polygon shapes
    polygons = [shape(polygon) for polygon in polygons]

    # Create a geopandas dataframe populated with the polygon shapes
    attribute_name = attribute_col if attribute_col is not None else "attribute"
    gdf = gpd.GeoDataFrame(
        data={attribute_name: values}, geometry=polygons, crs=da.odc.crs
    )

    # If a file path is supplied, export to file
    if output_path is not None:
        if verbose:
            print(f"Exporting vector data to {output_path}")
        gdf.to_file(output_path)

    return gdf


def xr_rasterize(
    gdf,
    da,
    attribute_col=None,
    crs=None,
    name=None,
    output_path=None,
    verbose=True,
    **rasterio_kwargs,
):
    """
    Rasterizes a vector ``geopandas.GeoDataFrame`` into a
    raster ``xarray.DataArray``.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        A ``geopandas.GeoDataFrame`` object containing the vector
        data you want to rasterise.
    da : xarray.DataArray or xarray.Dataset
        The shape, coordinates, dimensions, and transform of this object
        are used to define the array that ``gdf`` is rasterized into.
        It effectively provides a spatial template.
    attribute_col : string, optional
        Name of the attribute column in ``gdf`` containing values for
        each vector feature that will be rasterized. If None, the
        output will be a boolean array of 1's and 0's.
    crs : str or CRS object, optional
        If ``da``'s coordinate reference system (CRS) cannot be
        determined, provide a CRS using this parameter.
        (e.g. 'EPSG:3577').
    name : str, optional
        An optional name used for the output ``xarray.DataArray`.
    output_path : string, optional
        Provide an optional string file path to export the rasterized
        data as a GeoTIFF file.
    verbose : bool, optional
        Print debugging messages. Default True.
    **rasterio_kwargs :
        A set of keyword arguments to ``rasterio.features.rasterize``.
        Can include: 'all_touched', 'merge_alg', 'dtype'.

    Returns
    -------
    da_rasterized : xarray.DataArray
        The rasterized vector data.
    """

    # Add GeoBox and odc.* accessor to array using `odc-geo`
    da = add_geobox(da, crs)

    # Reproject vector data to raster's CRS
    gdf_reproj = gdf.to_crs(crs=da.odc.crs)

    # If an attribute column is specified, rasterise using vector
    # attribute values. Otherwise, rasterise into a boolean array
    if attribute_col is not None:
        # Use the geometry and attributes from `gdf` to create an iterable
        shapes = zip(gdf_reproj.geometry, gdf_reproj[attribute_col])
    else:
        # Use geometry directly (will produce a boolean numpy array)
        shapes = gdf_reproj.geometry

    # Rasterise shapes into a numpy array
    im = rasterio.features.rasterize(
        shapes=shapes,
        out_shape=da.odc.geobox.shape,
        transform=da.odc.geobox.transform,
        **rasterio_kwargs,
    )

    # Convert numpy array to a full xarray.DataArray
    # and set array name if supplied
    da_rasterized = odc.geo.xr.wrap_xr(im=im, gbox=da.odc.geobox)
    da_rasterized = da_rasterized.rename(name)

    # If a file path is supplied, export to file
    if output_path is not None:
        if verbose:
            print(f"Exporting raster data to {output_path}")
        write_cog(da_rasterized, output_path, overwrite=True)

    return da_rasterized


def subpixel_contours(
    da,
    z_values=[0.0],
    crs=None,
    attribute_df=None,
    output_path=None,
    min_vertices=2,
    dim="time",
    time_format="%Y-%m-%d",
    errors="ignore",
    verbose=True,
):
    """
    Uses `skimage.measure.find_contours` to extract multiple z-value
    contour lines from a two-dimensional array (e.g. multiple elevations
    from a single DEM), or one z-value for each array along a specified
    dimension of a multi-dimensional array (e.g. to map waterlines
    across time by extracting a 0 NDWI contour from each individual
    timestep in an xarray timeseries).

    Contours are returned as a geopandas.GeoDataFrame with one row per
    z-value or one row per array along a specified dimension. The
    `attribute_df` parameter can be used to pass custom attributes
    to the output contour features.

    Last modified: May 2023

    Parameters
    ----------
    da : xarray DataArray
        A two-dimensional or multi-dimensional array from which
        contours are extracted. If a two-dimensional array is provided,
        the analysis will run in 'single array, multiple z-values' mode
        which allows you to specify multiple `z_values` to be extracted.
        If a multi-dimensional array is provided, the analysis will run
        in 'single z-value, multiple arrays' mode allowing you to
        extract contours for each array along the dimension specified
        by the `dim` parameter.
    z_values : int, float or list of ints, floats
        An individual z-value or list of multiple z-values to extract
        from the array. If operating in 'single z-value, multiple
        arrays' mode specify only a single z-value.
    crs : string or CRS object, optional
        If ``da``'s coordinate reference system (CRS) cannot be
        determined, provide a CRS using this parameter.
        (e.g. 'EPSG:3577').
    output_path : string, optional
        The path and filename for the output shapefile.
    attribute_df : pandas.Dataframe, optional
        A pandas.Dataframe containing attributes to pass to the output
        contour features. The dataframe must contain either the same
        number of rows as supplied `z_values` (in 'multiple z-value,
        single array' mode), or the same number of rows as the number
        of arrays along the `dim` dimension ('single z-value, multiple
        arrays mode').
    min_vertices : int, optional
        The minimum number of vertices required for a contour to be
        extracted. The default (and minimum) value is 2, which is the
        smallest number required to produce a contour line (i.e. a start
        and end point). Higher values remove smaller contours,
        potentially removing noise from the output dataset.
    dim : string, optional
        The name of the dimension along which to extract contours when
        operating in 'single z-value, multiple arrays' mode. The default
        is 'time', which extracts contours for each array along the time
        dimension.
    time_format : string, optional
        The format used to convert `numpy.datetime64` values to strings
        if applied to data with a "time" dimension. Defaults to
        "%Y-%m-%d".
    errors : string, optional
        If 'raise', then any failed contours will raise an exception.
        If 'ignore' (the default), a list of failed contours will be
        printed. If no contours are returned, an exception will always
        be raised.
    verbose : bool, optional
        Print debugging messages. Default is True.

    Returns
    -------
    output_gdf : geopandas geodataframe
        A geopandas geodataframe object with one feature per z-value
        ('single array, multiple z-values' mode), or one row per array
        along the dimension specified by the `dim` parameter ('single
        z-value, multiple arrays' mode). If `attribute_df` was
        provided, these values will be included in the shapefile's
        attribute table.
    """

    def _contours_to_multiline(da_i, z_value, min_vertices=2):
        """
        Helper function to apply marching squares contour extraction
        to an array and return a data as a shapely MultiLineString.
        The `min_vertices` parameter allows you to drop small contours
        with less than X vertices.
        """

        # Extracts contours from array, and converts each discrete
        # contour into a Shapely LineString feature. If the function
        # returns a KeyError, this may be due to an unresolved issue in
        # scikit-image: https://github.com/scikit-image/scikit-image/issues/4830
        # A temporary workaround is to peturb the z-value by a tiny
        # amount (1e-12) before using it to extract the contour.
        try:
            line_features = [
                LineString(i[:, [1, 0]])
                for i in find_contours(da_i.data, z_value)
                if i.shape[0] >= min_vertices
            ]
        except KeyError:
            line_features = [
                LineString(i[:, [1, 0]])
                for i in find_contours(da_i.data, z_value + 1e-12)
                if i.shape[0] >= min_vertices
            ]

        # Output resulting lines into a single combined MultiLineString
        return MultiLineString(line_features)

    def _time_format(i, time_format):
        """
        Converts numpy.datetime64 into formatted strings;
        otherwise returns data as-is.
        """
        if isinstance(i, np.datetime64):
            ts = pd.to_datetime(str(i))
            i = ts.strftime(time_format)
        return i

    # Verify input data is a xr.DataArray
    if not isinstance(da, xr.DataArray):
        raise ValueError(
            "The input `da` is not an xarray.DataArray. "
            "If you supplied an xarray.Dataset, pass in one "
            "of its data variables using the syntax "
            "`da=ds.<variable name>`."
        )

    # Add GeoBox and odc.* accessor to array using `odc-geo`
    da = add_geobox(da, crs)

    # If z_values is supplied is not a list, convert to list:
    z_values = (
        z_values
        if (isinstance(z_values, list) or isinstance(z_values, np.ndarray))
        else [z_values]
    )

    # If dask collection, load into memory
    if dask.is_dask_collection(da):
        if verbose:
            print("Loading data into memory using Dask")
        da = da.compute()

    # Test number of dimensions in supplied data array
    if len(da.shape) == 2:
        if verbose:
            print("Operating in multiple z-value, single array mode")
        dim = "z_value"
        contour_arrays = {
            _time_format(i, time_format): _contours_to_multiline(da, i, min_vertices)
            for i in z_values
        }

    else:
        # Test if only a single z-value is given when operating in
        # single z-value, multiple arrays mode
        if verbose:
            print("Operating in single z-value, multiple arrays mode")
        if len(z_values) > 1:
            raise ValueError(
                "Please provide a single z-value when operating "
                "in single z-value, multiple arrays mode"
            )

        contour_arrays = {
            _time_format(i, time_format): _contours_to_multiline(
                da_i, z_values[0], min_vertices
            )
            for i, da_i in da.groupby(dim)
        }

    # If attributes are provided, add the contour keys to that dataframe
    if attribute_df is not None:
        try:
            attribute_df.insert(0, dim, contour_arrays.keys())

        # If this fails, it is due to the applied attribute table not
        # matching the structure of the loaded data
        except ValueError:
            if len(da.shape) == 2:
                raise ValueError(
                    f"The provided `attribute_df` contains a different "
                    f"number of rows ({len(attribute_df.index)}) "
                    f"than the number of supplied `z_values` "
                    f"({len(z_values)})."
                )
            else:
                raise ValueError(
                    f"The provided `attribute_df` contains a different "
                    f"number of rows ({len(attribute_df.index)}) "
                    f"than the number of arrays along the '{dim}' "
                    f"dimension ({len(da[dim])})."
                )

    # Otherwise, use the contour keys as the only main attributes
    else:
        attribute_df = list(contour_arrays.keys())

    # Convert output contours to a geopandas.GeoDataFrame
    contours_gdf = gpd.GeoDataFrame(
        data=attribute_df, geometry=list(contour_arrays.values()), crs=da.odc.crs
    )

    # Define affine and use to convert array coords to geographic coords.
    # We need to add 0.5 x pixel size to the x and y to obtain the centre
    # point of our pixels, rather than the top-left corner
    affine = da.odc.geobox.transform
    shapely_affine = [
        affine.a,
        affine.b,
        affine.d,
        affine.e,
        affine.xoff + affine.a / 2.0,
        affine.yoff + affine.e / 2.0,
    ]
    contours_gdf["geometry"] = contours_gdf.affine_transform(shapely_affine)

    # Rename the data column to match the dimension
    contours_gdf = contours_gdf.rename({0: dim}, axis=1)

    # Drop empty timesteps
    empty_contours = contours_gdf.geometry.is_empty
    failed = ", ".join(map(str, contours_gdf[empty_contours][dim].to_list()))
    contours_gdf = contours_gdf[~empty_contours]

    # Raise exception if no data is returned, or if any contours fail
    # when `errors='raise'. Otherwise, print failed contours
    if empty_contours.all() and errors == "raise":
        raise ValueError(
            "Failed to generate any valid contours; verify that "
            "values passed to `z_values` are valid and present "
            "in `da`"
        )
    elif empty_contours.all() and errors == "ignore":
        if verbose:
            print(
                "Failed to generate any valid contours; verify that "
                "values passed to `z_values` are valid and present "
                "in `da`"
            )
    elif empty_contours.any() and errors == "raise":
        raise Exception(f"Failed to generate contours: {failed}")
    elif empty_contours.any() and errors == "ignore":
        if verbose:
            print(f"Failed to generate contours: {failed}")

    # If asked to write out file, test if GeoJSON or ESRI Shapefile. If
    # GeoJSON, convert to EPSG:4326 before exporting.
    if output_path and output_path.endswith(".geojson"):
        if verbose:
            print(f"Writing contours to {output_path}")
        contours_gdf.to_crs("EPSG:4326").to_file(filename=output_path)

    if output_path and output_path.endswith(".shp"):
        if verbose:
            print(f"Writing contours to {output_path}")
        contours_gdf.to_file(filename=output_path)

    return contours_gdf


def interpolate_2d(ds,
                   x_coords,
                   y_coords,
                   z_coords,
                   method='linear',
                   factor=1,
                   verbose=False,
                   **kwargs):

    """
    This function takes points with X, Y and Z coordinates, and
    interpolates Z-values across the extent of an existing xarray
    dataset. This can be useful for producing smooth surfaces from point
    data that can be compared directly against satellite data derived
    from an OpenDataCube query.

    Supported interpolation methods include 'linear', 'nearest' and
    'cubic (using `scipy.interpolate.griddata`), and 'rbf' (using
    `scipy.interpolate.Rbf`).

    Last modified: February 2020

    Parameters
    ----------
    ds : xarray DataArray or Dataset
        A two-dimensional or multi-dimensional array from which x and y
        dimensions will be copied and used for the area in which to
        interpolate point data.
    x_coords, y_coords : numpy array
        Arrays containing X and Y coordinates for all points (e.g.
        longitudes and latitudes).
    z_coords : numpy array
        An array containing Z coordinates for all points (e.g.
        elevations). These are the values you wish to interpolate
        between.
    method : string, optional
        The method used to interpolate between point values. This string
        is either passed to `scipy.interpolate.griddata` (for 'linear',
        'nearest' and 'cubic' methods), or used to specify Radial Basis
        Function interpolation using `scipy.interpolate.Rbf` ('rbf').
        Defaults to 'linear'.
    factor : int, optional
        An optional integer that can be used to subsample the spatial
        interpolation extent to obtain faster interpolation times, then
        up-sample this array back to the original dimensions of the
        data as a final step. For example, setting `factor=10` will
        interpolate data into a grid that has one tenth of the
        resolution of `ds`. This approach will be significantly faster
        than interpolating at full resolution, but will potentially
        produce less accurate or reliable results.
    verbose : bool, optional
        Print debugging messages. Default False.
    **kwargs :
        Optional keyword arguments to pass to either
        `scipy.interpolate.griddata` (if `method` is 'linear', 'nearest'
        or 'cubic'), or `scipy.interpolate.Rbf` (is `method` is 'rbf').

    Returns
    -------
    interp_2d_array : xarray DataArray
        An xarray DataArray containing with x and y coordinates copied
        from `ds_array`, and Z-values interpolated from the points data.
    """

    # Extract xy and elev points
    points_xy = np.vstack([x_coords, y_coords]).T

    # Extract x and y coordinates to interpolate into.
    # If `factor` is greater than 1, the coordinates will be subsampled
    # for faster run-times. If the last x or y value in the subsampled
    # grid aren't the same as the last x or y values in the original
    # full resolution grid, add the final full resolution grid value to
    # ensure data is interpolated up to the very edge of the array
    if ds.x[::factor][-1].item() == ds.x[-1].item():
        x_grid_coords = ds.x[::factor].values
    else:
        x_grid_coords = ds.x[::factor].values.tolist() + [ds.x[-1].item()]

    if ds.y[::factor][-1].item() == ds.y[-1].item():
        y_grid_coords = ds.y[::factor].values
    else:
        y_grid_coords = ds.y[::factor].values.tolist() + [ds.y[-1].item()]

    # Create grid to interpolate into
    grid_y, grid_x = np.meshgrid(x_grid_coords, y_grid_coords)

    # Apply scipy.interpolate.griddata interpolation methods
    if method in ('linear', 'nearest', 'cubic'):

        # Interpolate x, y and z values
        interp_2d = scipy.interpolate.griddata(points=points_xy,
                                               values=z_coords,
                                               xi=(grid_y, grid_x),
                                               method=method,
                                               **kwargs)

    # Apply Radial Basis Function interpolation
    elif method == 'rbf':

        # Interpolate x, y and z values
        rbf = scipy.interpolate.Rbf(x_coords, y_coords, z_coords, **kwargs)
        interp_2d = rbf(grid_y, grid_x)

    # Create xarray dataarray from the data and resample to ds coords
    interp_2d_da = xr.DataArray(interp_2d,
                                coords=[y_grid_coords, x_grid_coords],
                                dims=['y', 'x'])

    # If factor is greater than 1, resample the interpolated array to
    # match the input `ds` array
    if factor > 1:
        interp_2d_da = interp_2d_da.interp_like(ds)

    return interp_2d_da


def contours_to_arrays(gdf, col):
    """
    This function converts a polyline shapefile into an array with three
    columns giving the X, Y and Z coordinates of each vertex. This data
    can then be used as an input to interpolation procedures (e.g. using
    a function like `interpolate_2d`.

    Last modified: October 2021

    Parameters
    ----------
    gdf : Geopandas GeoDataFrame
        A GeoPandas GeoDataFrame of lines to convert into point
        coordinates.
    col : str
        A string giving the name of the GeoDataFrame field to use as
        Z-values.

    Returns
    -------
    A numpy array with three columns giving the X, Y and Z coordinates
    of each vertex in the input GeoDataFrame.

    """

    # Explode multi-part geometries into multiple single geometries.
    gdf = gdf.explode(ignore_index=True)

    coords_zvals = []

    for i in range(0, len(gdf)):
        val = gdf.iloc[i][col]

        try:
            coords = np.concatenate(
                [np.vstack(x.coords.xy).T for x in gdf.iloc[i].geometry.geoms]
            )
        except Exception:
            coords = np.vstack(gdf.iloc[i].geometry.coords.xy).T

        coords_zvals.append(
            np.column_stack((coords, np.full(np.shape(coords)[0], fill_value=val)))
        )

    return np.concatenate(coords_zvals)


def largest_region(bool_array, **kwargs):

    '''
    Takes a boolean array and identifies the largest contiguous region of
    connected True values. This is returned as a new array with cells in
    the largest region marked as True, and all other cells marked as False.

    Parameters
    ----------
    bool_array : boolean array
        A boolean array (numpy or xarray.DataArray) with True values for
        the areas that will be inspected to find the largest group of
        connected cells
    **kwargs :
        Optional keyword arguments to pass to `measure.label`

    Returns
    -------
    largest_region : boolean array
        A boolean array with cells in the largest region marked as True,
        and all other cells marked as False.

    '''

    # First, break boolean array into unique, discrete regions/blobs
    blobs_labels = label(bool_array, background=0, **kwargs)

    # Count the size of each blob, excluding the background class (0)
    ids, counts = np.unique(blobs_labels[blobs_labels > 0],
                            return_counts=True)

    # Identify the region ID of the largest blob
    largest_region_id = ids[np.argmax(counts)]

    # Produce a boolean array where 1 == the largest region
    largest_region = blobs_labels == largest_region_id

    return largest_region


def transform_geojson_wgs_to_epsg(geojson, EPSG):
    """
    Takes a geojson dictionary and converts it from WGS84 (EPSG:4326) to desired EPSG

    Parameters
    ----------
    geojson: dict
        a geojson dictionary containing a 'geometry' key, in WGS84 coordinates
    EPSG: int
        numeric code for the EPSG coordinate referecnce system to transform into

    Returns
    -------
    transformed_geojson: dict
        a geojson dictionary containing a 'coordinates' key, in the desired CRS

    """
    gg = Geometry(geojson['geometry'], CRS('epsg:4326'))
    gg = gg.to_crs(CRS(f'epsg:{EPSG}'))
    return gg.__geo_interface__


def zonal_stats_parallel(shp,
                         raster,
                         statistics,
                         out_shp,
                         ncpus,
                         **kwargs):

    """
    Summarizing raster datasets based on vector geometries in parallel.
    Each cpu recieves an equal chunk of the dataset.
    Utilizes the perrygeo/rasterstats package.

    Parameters
    ----------
    shp : str
        Path to shapefile that contains polygons over
        which zonal statistics are calculated
    raster: str
        Path to the raster from which the statistics are calculated.
        This can be a virtual raster (.vrt).
    statistics: list
        list of statistics to calculate. e.g.
            ['min', 'max', 'median', 'majority', 'sum']
    out_shp: str
        Path to export shapefile containing zonal statistics.
    ncpus: int
        number of cores to parallelize the operations over.
    kwargs:
        Any other keyword arguments to rasterstats.zonal_stats()
        See https://github.com/perrygeo/python-rasterstats for
        all options

    Returns
    -------
    Exports a shapefile to disk containing the zonal statistics requested

    """

    # yields n sized chunks from list l (used for splitting task to multiple processes)
    def chunks(l, n):
        for i in range(0, len(l), n):
            yield l[i:i + n]

    # calculates zonal stats and adds results to a dictionary
    def worker(z, raster, d):
        z_stats = zonal_stats(z, raster, stats=statistics, **kwargs)
        for i in range(0, len(z_stats)):
            d[z[i]['id']] = z_stats[i]

    # write output polygon
    def write_output(zones, out_shp, d):
        # copy schema and crs from input and add new fields for each statistic
        schema = zones.schema.copy()
        crs = zones.crs
        for stat in statistics:
            schema['properties'][stat] = 'float'

        with fiona.open(out_shp, 'w', 'ESRI Shapefile', schema, crs) as output:
            for elem in zones:
                for stat in statistics:
                    elem['properties'][stat] = d[elem['id']][stat]
                output.write({'properties': elem['properties'], 'geometry': mapping(shape(elem['geometry']))})

    with fiona.open(shp) as zones:
        jobs = []

        # create manager dictionary (polygon ids=keys, stats=entries)
        # where multiple processes can write without conflicts
        man = mp.Manager()
        d = man.dict()

        # split zone polygons into 'ncpus' chunks for parallel processing
        # and call worker() for each
        split = chunks(zones, len(zones)//ncpus)
        for z in split:
            p = mp.Process(target=worker, args=(z, raster, d))
            p.start()
            jobs.append(p)

        # wait that all chunks are finished
        [j.join() for j in jobs]

        write_output(zones, out_shp, d)


def reverse_geocode(coords, site_classes=None, state_classes=None):
    """
    Takes a latitude and longitude coordinate, and performs a reverse
    geocode to return a plain-text description of the location in the
    form:

        Site, State

    E.g.: `reverse_geocode(coords=(-35.282163, 149.128835))`

        'Canberra, Australian Capital Territory'

    Parameters
    ----------
    coords : tuple of floats
        A tuple of (latitude, longitude) coordinates used to perform
        the reverse geocode.
    site_classes : list of strings, optional
        A list of strings used to define the site part of the plain
        text location description. Because the contents of the geocoded
        address can vary greatly depending on location, these strings
        are tested against the address one by one until a match is made.

        Defaults to:

            ``['city', 'town', 'village', 'suburb', 'hamlet', 'county', 'municipality']``

    state_classes : list of strings, optional
        A list of strings used to define the state part of the plain
        text location description. These strings are tested against the
        address one by one until a match is made. Defaults to:
        `['state', 'territory']`.
    Returns
    -------
    If a valid geocoded address is found, a plain text location
    description will be returned:

        'Site, State'

    If no valid address is found, formatted coordinates will be returned
    instead:

        'XX.XX S, XX.XX E'
    """

    # Run reverse geocode using coordinates
    geocoder = Nominatim(user_agent='Digital Earth Africa')
    out = geocoder.reverse(coords)

    # Create plain text-coords as fall-back
    lat = f'{-coords[0]:.2f} S' if coords[0] < 0 else f'{coords[0]:.2f} N'
    lon = f'{-coords[1]:.2f} W' if coords[1] < 0 else f'{coords[1]:.2f} E'

    try:

        # Get address from geocoded data
        address = out.raw['address']

        # Use site and state classes if supplied; else use defaults
        default_site_classes = ['city', 'town', 'village', 'suburb', 'hamlet',
                                'county', 'municipality']
        default_state_classes = ['state', 'territory']
        site_classes = site_classes if site_classes else default_site_classes
        state_classes = state_classes if state_classes else default_state_classes

        # Return the first site or state class that exists in address dict
        site = next((address[k] for k in site_classes if k in address), None)
        state = next((address[k] for k in state_classes if k in address), None)

        # If site and state exist in the data, return this.
        # Otherwise, return N/E/S/W coordinates.
        if site and state:

            # Return as site, state formatted string
            return f'{site}, {state}'

        else:

            # If no geocoding result, return N/E/S/W coordinates
            print('No valid geocoded location; returning coordinates instead')
            return f'{lat}, {lon}'

    except (KeyError, AttributeError):

        # If no geocoding result, return N/E/S/W coordinates
        print('No valid geocoded location; returning coordinates instead')
        return f'{lat}, {lon}'


def sun_angles(dc, query):
    """
    For a given spatiotemporal query, calculate mean sun
    azimuth and elevation for each satellite observation, and
    return these as a new `xarray.Dataset` with 'sun_elevation'
    and 'sun_azimuth' variables.

    Parameters:
    -----------
    dc : datacube.Datacube object
        Datacube instance used to load data.
    query : dict
        A dictionary containing query parameters used to identify
        satellite observations and load metadata.

    Returns:
    --------
    sun_angles_ds : xarray.Dataset
        An `xarray.set` containing a 'sun_elevation' and
        'sun_azimuth' variables.
    """
    # Identify satellite datasets and group outputs using the
    # same approach used to group satellite imagery (i.e. solar day)
    gb = query_group_by(**query)
    datasets = dc.find_datasets(**query)
    dataset_array = dc.group_datasets(datasets, gb)

    # Load and take the mean of metadata from each product
    sun_azimuth = xr_apply(
        dataset_array,
        lambda t, dd: np.mean([d.metadata.eo_sun_azimuth for d in dd]),
        dtype=float,
    )
    sun_elevation = xr_apply(
        dataset_array,
        lambda t, dd: np.mean([d.metadata.eo_sun_elevation for d in dd]),
        dtype=float,
    )

    # Combine into new xarray.Dataset
    sun_angles_ds = xr.merge(
        [sun_elevation.rename("sun_elevation"), sun_azimuth.rename("sun_azimuth")]
    )

    return sun_angles_ds



import xarray as xr
import numpy as np

# function to load soil moisture data

def load_soil_moisture(lat, lon, time, product = 'surface', grid = 'nearest'):
    product_baseurl = 'https://dapds00.nci.org.au/thredds/dodsC/ub8/global/GRAFS/'
    assert product in ['surface', 'rootzone'], 'product parameter must be surface or root-zone'
    # lat, lon grid
    if grid == 'nearest':
        # select lat/lon range from data; snap to nearest grid
        lat_range, lon_range = None, None
    else:
        # define a grid that covers the entire area of interest
        lat_range = np.arange(np.max(np.ceil(np.array(lat)*10.+0.5)/10.-0.05), np.min(np.floor(np.array(lat)*10.-0.5)/10.+0.05)-0.05, -0.1)
        lon_range = np.arange(np.min(np.floor(np.array(lon)*10.-0.5)/10.+0.05), np.max(np.ceil(np.array(lon)*10.+0.5)/10.-0.05)+0.05, 0.1)
    # split time window into years
    day_range = np.array(time).astype("M8[D]")
    year_range = np.array(time).astype("M8[Y]")
    if product == 'surface':
        product_name = 'GRAFS_TopSoilRelativeWetness_'
    else: product_name = 'GRAFS_RootzoneSoilWaterIndex_'
    datasets = []
    for year in np.arange(year_range[0], year_range[1]+1, np.timedelta64(1, 'Y')):
        start = np.max([day_range[0], year.astype("M8[D]")])
        end = np.min([day_range[1], (year+1).astype("M8[D]")-1])
        product_url = product_baseurl + product_name +'%s.nc'%str(year)
        print(product_url)
        # data is loaded lazily through OPeNDAP
        ds = xr.open_dataset(product_url)
        if lat_range is None:
            # select lat/lon range from data if not specified; snap to nearest grid
            test = ds.sel(lat=list(lat), lon=list(lon), method='nearest')
            lat_range = slice(test.lat.values[0], test.lat.values[1])
            lon_range = slice(test.lon.values[0], test.lon.values[1])
        # slice before return
        ds = ds.sel(lat=lat_range, lon=lon_range, time=slice(start, end)).compute()
        datasets.append(ds)
    return xr.merge(datasets)


"""
Loading and interacting with data in the change filmstrips notebook,
inside the Real_world_examples folder.
"""

# Load modules
import os
import dask
import datacube
import warnings
import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt
from odc.algo import geomedian_with_mads
from odc.ui import select_on_a_map
from dask.utils import parse_bytes
from datacube.utils.geometry import CRS, assign_crs
from datacube.utils.rio import configure_s3_access
from datacube.utils.dask import start_local_dask
from ipyleaflet import basemaps, basemap_to_tiles

# Load utility functions
from deafrica_tools.datahandling import load_ard, mostcommon_crs
from deafrica_tools.dask import create_local_dask_cluster


def run_filmstrip_app(
        output_name,
        time_range,
        time_step,
        tide_range=(0.0, 1.0),
        resolution=(-30, 30),
        max_cloud=0.5,
        ls7_slc_off=False,
        size_limit=10000,
):
    """
    An interactive app that allows the user to select a region from a
    map, then load Digital Earth Africa Landsat data and combine it
    using the geometric median ("geomedian") statistic to reveal the
    median or 'typical' appearance of the landscape for a series of
    time periods.

    The results for each time period are combined into a 'filmstrip'
    plot which visualises how the landscape has changed in appearance
    across time, with a 'change heatmap' panel highlighting potential
    areas of greatest change.

    For coastal applications, the analysis can be customised to select
    only satellite images obtained during a specific tidal range
    (e.g. low, average or high tide).

    Last modified: April 2020

    Parameters
    ----------
    output_name : str
        A name that will be used to name the output filmstrip plot file.
    time_range : tuple
        A tuple giving the date range to analyse
        (e.g. `time_range = ('1988-01-01', '2017-12-31')`).
    time_step : dict
        This parameter sets the length of the time periods to compare
        (e.g. `time_step = {'years': 5}` will generate one filmstrip
        plot for every five years of data; `time_step = {'months': 18}`
        will generate one plot for each 18 month period etc. Time
        periods are counted from the first value given in `time_range`.
    tide_range : tuple, optional
        An optional parameter that can be used to generate filmstrip
        plots based on specific ocean tide conditions. This can be
        valuable for analysing change consistently along the coast.
        For example, `tide_range = (0.0, 0.2)` will select only
        satellite images acquired at the lowest 20% of tides;
        `tide_range = (0.8, 1.0)` will select images from the highest
        20% of tides. The default is `tide_range = (0.0, 1.0)` which
        will select all images regardless of tide.
    resolution : tuple, optional
        The spatial resolution to load data. The default is
        `resolution = (-30, 30)`, which will load data at 30 m pixel
        resolution. Increasing this (e.g. to `resolution = (-100, 100)`)
        can be useful for loading large spatial extents.
    max_cloud : float, optional
        This parameter can be used to exclude satellite images with
        excessive cloud. The default is `0.5`, which will keep all images
        with less than 50% cloud.
    ls7_slc_off : bool, optional
        An optional boolean indicating whether to include data from
        after the Landsat 7 SLC failure (i.e. SLC-off). Defaults to
        False, which removes all Landsat 7 observations > May 31 2003.
    size_limit : int, optional
        An optional integer (in hectares) specifying the size limit
        for the data query. Queries larger than this size will receive
        a warning that he data query is too large (and may
        therefore result in memory errors).


    Returns
    -------
    ds_geomedian : xarray Dataset
        An xarray dataset containing geomedian composites for each
        timestep in the analysis.

    """

    ########################
    # Select and load data #
    ########################

    # Define centre_coords as a global variable
    global centre_coords

    # Test if centre_coords is in the global namespace;
    # use default value if it isn't
    if "centre_coords" not in globals():
        centre_coords = (6.587292, 1.532833)

    # Plot interactive map to select area
    basemap = basemap_to_tiles(basemaps.Esri.WorldImagery)
    geopolygon = select_on_a_map(height="600px",
                                 layers=(basemap,),
                                 center=centre_coords,
                                 zoom=14)

    # Set centre coords based on most recent selection to re-focus
    # subsequent data selections
    centre_coords = geopolygon.centroid.points[0][::-1]

    # Test size of selected area
    msq_per_hectare = 10000
    area = geopolygon.to_crs(crs=CRS("epsg:6933")).area / msq_per_hectare
    radius = np.round(np.sqrt(size_limit), 1)
    if area > size_limit:
        print(f"Warning: Your selected area is {area:.00f} hectares. "
              f"Please select an area of less than {size_limit} hectares."
              f"\nTo select a smaller area, re-run the cell "
              f"above and draw a new polygon.")

    else:

        print("Starting analysis...")

        # Connect to datacube database
        dc = datacube.Datacube(app="Change_filmstrips")

        # Configure local dask cluster
        client = create_local_dask_cluster(return_client=True)

        # Obtain native CRS
        crs = mostcommon_crs(dc=dc,
                             product="ls8_sr",
                             query={
                                 "time": "2014",
                                 "geopolygon": geopolygon
                             })

        # Create query based on time range, area selected, custom params
        query = {
            "time": time_range,
            "geopolygon": geopolygon,
            "output_crs": crs,
            "resolution": resolution,
            "dask_chunks": {
                "x": 3000,
                "y": 3000
            },
            "align": (resolution[1] / 2.0, resolution[1] / 2.0),
        }

        # Load data from all three Landsats
        warnings.filterwarnings("ignore")
        ds = load_ard(
            dc=dc,
            measurements=["red", "green", "blue"],
            products=["ls5_sr", "ls7_sr", "ls8_sr"],
            min_gooddata=max_cloud,
            ls7_slc_off=ls7_slc_off,
            **query,
        )

        # Optionally calculate tides for each timestep in the satellite
        # dataset and drop any observations out side this range
        if tide_range != (0.0, 1.0):
            from deafrica_tools.coastal import tidal_tag
            ds = tidal_tag(ds=ds, tidepost_lat=None, tidepost_lon=None)
            min_tide, max_tide = ds.tide_height.quantile(tide_range).values
            ds = ds.sel(time=(ds.tide_height >= min_tide) &
                        (ds.tide_height <= max_tide))
            ds = ds.drop("tide_height")
            print(f"    Keeping {len(ds.time)} observations with tides "
                  f"between {min_tide:.2f} and {max_tide:.2f} m")

        # Create time step ranges to generate filmstrips from
        bins_dt = pd.date_range(start=time_range[0],
                                end=time_range[1],
                                freq=pd.DateOffset(**time_step))

        # Bin all satellite observations by timestep. If some observations
        # fall outside the upper bin, label these with the highest bin
        labels = bins_dt.astype("str")
        time_steps = (pd.cut(ds.time.values, bins_dt,
                             labels=labels[:-1]).add_categories(
                                 labels[-1]).fillna(labels[-1]))

        time_steps_var = xr.DataArray(time_steps, [("time", ds.time.values)],
                                      name="timestep")

        # Resample data temporally into time steps, and compute geomedians
        ds_geomedian = (ds.groupby(time_steps_var).apply(
            lambda ds_subset: geomedian_with_mads(
                ds_subset, compute_mads=False, compute_count=False)))

        print("\nGenerating geomedian composites and plotting "
              "filmstrips... (click the Dashboard link above for status)")
        ds_geomedian = ds_geomedian.compute()

        # Reset CRS that is lost during geomedian compositing
        ds_geomedian = assign_crs(ds_geomedian, crs=ds.geobox.crs)

        ############
        # Plotting #
        ############

        # Convert to array and extract vmin/vmax
        output_array = ds_geomedian[["red", "green", "blue"]].to_array()
        percentiles = output_array.quantile(q=(0.02, 0.98)).values

        # Create the plot with one subplot more than timesteps in the
        # dataset. Figure width is set based on the number of subplots
        # and aspect ratio
        n_obs = output_array.sizes["timestep"]
        ratio = output_array.sizes["x"] / output_array.sizes["y"]
        fig, axes = plt.subplots(1,
                                 n_obs + 1,
                                 figsize=(5 * ratio * (n_obs + 1), 5))
        fig.subplots_adjust(wspace=0.05, hspace=0.05)

        # Add timesteps to the plot, set aspect to equal to preserve shape
        for i, ax_i in enumerate(axes.flatten()[:n_obs]):
            output_array.isel(timestep=i).plot.imshow(ax=ax_i,
                                                      vmin=percentiles[0],
                                                      vmax=percentiles[1])
            ax_i.get_xaxis().set_visible(False)
            ax_i.get_yaxis().set_visible(False)
            ax_i.set_aspect("equal")

        # Add change heatmap panel to final subplot. Heatmap is computed
        # by first taking the log of the array (so change in dark areas
        # can be identified), then computing standard deviation between
        # all timesteps
        (np.log(output_array).std(dim=["timestep"]).mean(
            dim="variable").plot.imshow(ax=axes.flatten()[-1],
                                        robust=True,
                                        cmap="magma",
                                        add_colorbar=False))
        axes.flatten()[-1].get_xaxis().set_visible(False)
        axes.flatten()[-1].get_yaxis().set_visible(False)
        axes.flatten()[-1].set_aspect("equal")
        axes.flatten()[-1].set_title("Change heatmap")

        # Export to file
        date_string = "_".join(time_range)
        ts_v = list(time_step.values())[0]
        ts_k = list(time_step.keys())[0]
        fig.savefig(
            f"filmstrip_{output_name}_{date_string}_{ts_v}{ts_k}.png",
            dpi=150,
            bbox_inches="tight",
            pad_inches=0.1,
        )

        # close dask client
        client.shutdown()

        return ds_geomedian



"""
Functions for simplifying the creation of a local dask cluster.
"""

from importlib.util import find_spec
import os
import dask
from aiohttp import ClientConnectionError
from datacube.utils.dask import start_local_dask
from datacube.utils.rio import configure_s3_access

_HAVE_PROXY = bool(find_spec('jupyter_server_proxy'))
_IS_AWS = ('AWS_ACCESS_KEY_ID' in os.environ or
           'AWS_DEFAULT_REGION' in os.environ)


def create_local_dask_cluster(spare_mem='3Gb', display_client=True, return_client=False):
    """
    Using the datacube utils function `start_local_dask`, generate
    a local dask cluster. Automatically detects if on AWS or NCI.

    Parameters
    ----------
    spare_mem : String, optional
        The amount of memory, in Gb, to leave for the notebook to run.
        This memory will not be used by the cluster. e.g '3Gb'
    display_client : Bool, optional
        An optional boolean indicating whether to display a summary of
        the dask client, including a link to monitor progress of the
        analysis. Set to False to hide this display.
    return_client : Bool, optional
        An optional boolean indicating whether to return the dask client
        object.

    """

    if _HAVE_PROXY:
        # Configure dashboard link to go over proxy
        prefix = os.environ.get('JUPYTERHUB_SERVICE_PREFIX', '/')
        dask.config.set({"distributed.dashboard.link":
                         prefix + "proxy/{port}/status"})

    # Start up a local cluster
    client = start_local_dask(mem_safety_margin=spare_mem)

    if _IS_AWS:
        # Configure GDAL for s3 access
        configure_s3_access(aws_unsigned=True,
                            client=client)

    # Show the dask cluster settings
    if display_client:
        from IPython.display import display
        display(client)
    
    # return the client as an object
    if return_client:
        return client
   

try:
    from dask_gateway import Gateway

    def create_dask_gateway_cluster(profile='r5_L', workers=2):
        """
        Create a cluster in our internal dask cluster.

        Parameters
        ----------
        profile : str
            Possible values are:
                - r5_L (2 cores, 15GB memory)
                - r5_XL (4 cores, 31GB memory)
                - r5_2XL (8 cores, 63GB memory)
                - r5_4XL (16 cores, 127GB memory)

        workers : int
            Number of workers in the cluster.
        """
        try:
            gateway = Gateway()
            
            # Close any existing clusters
            cluster_names = gateway.list_clusters()
            if len(cluster_names) > 0:
                print("Cluster(s) still running:", cluster_names)
                for n in cluster_names:
                    cluster = gateway.connect(n.name)
                    cluster.shutdown()            
            
            options = gateway.cluster_options()
            options['profile'] = profile

            # limit username to alphanumeric characters
            # kubernetes pods won't launch if labels contain anything other than [a-Z, -, _]
            options['jupyterhub_user'] = ''.join(c if c.isalnum() else '-' for c in os.getenv('JUPYTERHUB_USER'))

            cluster = gateway.new_cluster(options)
            cluster.scale(workers)
            return cluster
        except ClientConnectionError:
            raise ConnectionError("access to dask gateway cluster unauthorized")

except ImportError:
    def create_dask_gateway_cluster(*args, **kwargs):
        raise NotImplementedError


"""
Create an interactive map for selecting satellite imagery and exporting image files.
"""

# Load modules
import datacube
import itertools
import numpy as np
import matplotlib.pyplot as plt
from odc.ui import select_on_a_map
from datacube.utils.geometry import CRS
from datacube.utils import masking
from skimage import exposure
from ipyleaflet import (WMSLayer, basemaps, basemap_to_tiles)
from traitlets import Unicode

from deafrica_tools.spatial import reverse_geocode
from deafrica_tools.dask import create_local_dask_cluster


def select_region_app(date,
                      satellites,
                      size_limit=10000):
    """
    An interactive app that allows the user to select a region from a
    map using imagery from Sentinel-2 and Landsat. The output of this
    function is used as the input to :func:`export_image_app` to export high-
    resolution satellite images.

    Last modified: September 2021

    Parameters
    ----------
    date : str
        The exact date used to plot imagery on the interactive map
        (e.g. ``date='1988-01-01'``).
    satellites : str
        The satellite data to plot on the interactive map. The 
        following options are supported:

            ``'Landsat-9'``: data from the Landsat 9 satellite
            ``'Landsat-8'``: data from the Landsat 8 satellite
            ``'Landsat-7'``: data from the Landsat 7 satellite
            ``'Landsat-5'``: data from the Landsat 5 satellite
            ``'Sentinel-2'``: data from Sentinel-2A and Sentinel-2B
            ``'Sentinel-2 geomedian'``: data from the Sentinel-2 annual geomedian

    size_limit : int, optional
        An optional size limit for the area selection in sq km.
        Defaults to 10000 sq km.

    Returns
    -------
    A dictionary containing:

        * 'geopolygon' (defining the area to export imagery from),
        * 'date' (date used to export imagery), and
        * 'satellites' (the satellites from which to extract imagery).

    These are passed to the :func:`export_image_app` function to export the image.
    """

    ########################
    # Select and load data #
    ########################

    # Load DEA WMS
    class TimeWMSLayer(WMSLayer):
        time = Unicode('').tag(sync=True, o=True)

    # WMS layers
    wms_params = {
        'Landsat-9': 'ls9_sr',
        'Landsat-8': 'ls8_sr',
        'Landsat-7': 'ls7_sr',
        'Landsat-5': 'ls5_sr',
        'Sentinel-2': 's2_l2a',
        'Sentinel-2 geomedian': 'gm_s2_annual'
    }

    time_wms = TimeWMSLayer(url='https://ows.digitalearth.africa/',
                            layers=wms_params[satellites],
                            time=date,
                            format='image/png',
                            transparent=True,
                            attribution='Digital Earth Africa')

    # Plot interactive map to select area
    basemap = basemap_to_tiles(basemaps.OpenStreetMap.Mapnik)
    geopolygon = select_on_a_map(height='1000px',
                                 layers=(
                                     basemap,
                                     time_wms,
                                 ),
                                 center=(4, 20),
                                 zoom=4)

    # Test size of selected area
    area = geopolygon.to_crs(crs=CRS('epsg:6933')).area / 1000000
    if area > size_limit:
        print(f'Warning: Your selected area is {area:.00f} sq km. '
              f'Please select an area of less than {size_limit} sq km.'
              f'\nTo select a smaller area, re-run the cell '
              f'above and draw a new polygon.')

    else:
        return {'geopolygon': geopolygon,
                'date': date,
                'satellites': satellites}


def export_image_app(geopolygon,
                     date,
                     satellites,
                     style='True colour',
                     resolution=None,
                     vmin=0,
                     vmax=2000,
                     percentile_stretch=None,
                     power=None,
                     image_proc_funcs=None,
                     output_format="jpg",
                     standardise_name=False):
    """
    Exports Digital Earth Africa satellite data as an image file
    based on the extent and time period selected using
    :func:`select_region_app`. The function supports Sentinel-2 and Landsat
    data, creating True and False colour images.

    By default, files are named using:

        ``"<product> - <YYYY-MM-DD> - <site, state> - <description>.png"``

    Set ``standardise_name=True`` for a machine-readable name:

        ``"<product>_<YYYY-MM-DD>_<site-state>_<description>.png"``

    Last modified: September 2021

    Parameters
    ----------
    geopolygon : datacube.utils.geometry object
        A datacube geopolygon providing the spatial bounds used to load
        satellite data.
    date : str
        The exact date used to extract imagery
        (e.g. `date='1988-01-01'`).
    satellites : str
        The satellite data to be used to extract imagery. The 
        following options are supported:

            ``'Landsat-9'``: data from the Landsat 9 satellite
            ``'Landsat-8'``: data from the Landsat 8 satellite
            ``'Landsat-7'``: data from the Landsat 7 satellite
            ``'Landsat-5'``: data from the Landsat 5 satellite
            ``'Sentinel-2'``: data from Sentinel-2A and Sentinel-2B
            ``'Sentinel-2 geomedian'``: data from the Sentinel-2 annual geomedian

    style : str, optional
        The style used to produce the image. Two options are currently
        supported:

            * ``'True colour'``: Creates a true colour image using the red,
              green and blue satellite bands
            * ``'False colour'``: Creates a false colour image using
              short-wave infrared, infrared and green satellite bands.
              The specific bands used vary between Landsat and Sentinel-2.

    resolution : tuple, optional
        The spatial resolution to load data. By default, the tool will
        automatically set the best possible resolution depending on the
        satellites selected (i.e 30 m for Landsat, 10 m for Sentinel-2).
        Increasing this (e.g. to ``resolution=(-100, 100)``) can be useful
        for loading large spatial extents.
    vmin, vmax : int or float
        The minimum and maximum surface reflectance values used to
        clip the resulting imagery to enhance contrast.
    percentile_stretch : tuple of floats, optional
        An tuple of two floats (between 0.00 and 1.00) that can be used
        to clip the imagery to based on percentiles to get more control
        over the brightness and contrast of the image. The default is
        ``None``; ``(0.02, 0.98)`` is equivelent to ``robust=True``. If this
        parameter is used, ``vmin`` and ``vmax`` will have no effect.
    power : float, optional
        Raises imagery by a power to reduce bright features and
        enhance dark features. This can add extra definition over areas
        with extremely bright features like snow, beaches or salt pans.
    image_proc_funcs : list of funcs, optional
        An optional list containing functions that will be applied to
        the output image. This can include image processing functions
        such as increasing contrast, unsharp masking, saturation etc.
        The function should take AND return a `numpy.ndarray` with
        shape ``[y, x, bands]``. If your function has parameters, you
        can pass in custom values using a lambda function, e.g.:
        ``[lambda x: skimage.filters.unsharp_mask(x, radius=5, amount=0.2)]``
    output_format : str, optional
        The output file format of the image. Valid options include ``'jpg'``
        and ``'png'``. Defaults to ``'jpg'``.
    standardise_name : bool, optional
        Whether to export the image file with a machine-readable
        file name (e.g. ``<product>_<YYYY-MM-DD>_<site-state>_<description>.png``)
    """

    ###########################
    # Set up satellite params #
    ###########################

    sat_params = {
        'Landsat-9': {
            'products': ['ls9_sr'],
            'resolution': [-30, 30],
            'styles': {
                'True colour': ['red', 'green', 'blue'],
                'False colour': ['swir_1', 'nir', 'green']
            }
        },
        'Landsat-8': {
            'products': ['ls8_sr'],
            'resolution': [-30, 30],
            'styles': {
                'True colour': ['red', 'green', 'blue'],
                'False colour': ['swir_1', 'nir', 'green']
            }
        },
        'Landsat-7': {
            'products': ['ls7_sr'],
            'resolution': [-30, 30],
            'styles': {
                'True colour': ['red', 'green', 'blue'],
                'False colour': ['swir_1', 'nir', 'green']
            }
        },
        'Landsat-5': {
            'products': ['ls5_sr'],
            'resolution': [-30, 30],
            'styles': {
                'True colour': ['red', 'green', 'blue'],
                'False colour': ['swir_1', 'nir', 'green']
            }
        },
        'Sentinel-2': {
            'products': ['s2_l2a'],
            'resolution': [-10, 10],
            'styles': {
                'True colour': ['red', 'green', 'blue'],
                'False colour': ['swir_2', 'nir_1', 'green']
            }
        },
        'Sentinel-2 geomedian': {
            'products': ['gm_s2_annual'],
            'resolution': [-10, 10],
            'styles': {
                'True colour': ['red', 'green', 'blue'],
                'False colour': ['swir_2', 'nir_1', 'green']
            }
        },
    }

    #############
    # Load data #
    #############

    # Connect to datacube database
    dc = datacube.Datacube(app='Exporting_satellite_images')

    # Configure local dask cluster
    client = create_local_dask_cluster(return_client=True)

    # Create query after adjusting interval time to UTC by
    # adding a UTC offset of -10 hours.
    start_date = np.datetime64(date)
    query_params = {
        'time': (str(start_date)),
        'geopolygon': geopolygon
    }

    # Find matching datasets
    dss = [
        dc.find_datasets(product=i, **query_params)
        for i in sat_params[satellites]['products']
    ]
    dss = list(itertools.chain.from_iterable(dss))

    # Get CRS and sensor
    crs = str(dss[0].crs)

    if satellites == 'Sentinel-2 geomedian':
        sensor = satellites
    else:
        sensor = dss[0].metadata_doc['properties']['eo:platform'].capitalize()
        sensor = sensor[0:-1].replace('_', '-') + sensor[-1].capitalize()

    # Use resolution if provided, otherwise use default
    if resolution:
        sat_params[satellites]['resolution'] = resolution

    load_params = {
        'output_crs': crs,
        'resolution': sat_params[satellites]['resolution'],
        'resampling': 'bilinear'
    }

    # Load data from datasets
    ds = dc.load(datasets=dss,
                 measurements=sat_params[satellites]['styles'][style],
                 group_by='solar_day',
                 dask_chunks={
                     'time': 1,
                     'x': 3000,
                     'y': 3000
                 },
                 **load_params,
                 **query_params)
    ds = masking.mask_invalid_data(ds)

    rgb_array = ds.isel(time=0).to_array().values

    ############
    # Plotting #
    ############

    # Create unique file name
    centre_coords = geopolygon.centroid.coords[0][::-1]
    site = reverse_geocode(coords=centre_coords)
    fname = (f"{sensor} - {date} - {site} - {style}, "
             f"{load_params['resolution'][1]} m resolution.{output_format}")

    # Remove spaces and commas if requested
    if standardise_name:
        fname = fname.replace(' - ', '_').replace(', ',
                                                  '-').replace(' ',
                                                               '-').lower()

    print(
        f'\nExporting image to {fname}.\nThis may take several minutes to complete...'
    )

    # Convert to numpy array
    rgb_array = np.transpose(rgb_array, axes=[1, 2, 0])

    # If percentile stretch is supplied, calculate vmin and vmax
    # from percentiles
    if percentile_stretch:
        vmin, vmax = np.nanpercentile(rgb_array, percentile_stretch)

    # Raise by power to dampen bright features and enhance dark.
    # Raise vmin and vmax by same amount to ensure proper stretch
    if power:
        rgb_array = rgb_array**power
        vmin, vmax = vmin**power, vmax**power

    # Rescale/stretch imagery between vmin and vmax
    rgb_rescaled = exposure.rescale_intensity(rgb_array.astype(float),
                                              in_range=(vmin, vmax),
                                              out_range=(0.0, 1.0))

    # Apply image processing funcs
    if image_proc_funcs:
        for i, func in enumerate(image_proc_funcs):
            print(f'Applying custom function {i + 1}')
            rgb_rescaled = func(rgb_rescaled)

    # Plot RGB
    plt.imshow(rgb_rescaled)

    # Export to file
    plt.imsave(fname=fname, arr=rgb_rescaled, format=output_format)

    # Close dask client
    client.shutdown()

    print('Finished exporting image.')



"""
Loading and processing DE Africa Water Bodies data.
Last modified: November 2023
"""

# Import required packages
import geopandas as gpd
from owslib.wfs import WebFeatureService
from owslib.fes import PropertyIsEqualTo
from owslib.etree import etree
import pandas as pd

# URL for the DE Africa Water Bodies data on Dev Geoserver.
WFS_ADDRESS = "https://geoserver.digitalearth.africa/geoserver/wfs"
WFS_LAYER = "waterbodies:DEAfrica_Waterbodies"

def get_waterbody(geohash: str) -> gpd.GeoDataFrame:
    """Gets a waterbody polygon and metadata by geohash.
    
    Parameters
    ----------
    geohash : str
        The geohash/UID for a waterbody in DE Africa Water Bodies.
    
    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame with the polygon.
    """
    
    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    filter_ = PropertyIsEqualTo(propertyname="UID", literal=geohash)
    filterxml = etree.tostring(filter_.toXML()).decode("utf-8")
    response = wfs.getfeature(
        typename=WFS_LAYER,
        filter=filterxml,
        outputFormat="json",
    )
    wb_gpd = gpd.read_file(response)
    return wb_gpd


def get_waterbodies(bbox: tuple, crs="EPSG:4326") -> gpd.GeoDataFrame:
    """Gets the polygons and metadata for multiple water bodies by bbox.
    
    Parameters
    ----------
    bbox : (xmin, ymin, xmax, ymax)
        Bounding box.
    crs : str
        Optional CRS for the bounding box.
    
    Returns
    -------
    gpd.GeoDataFrame
        A GeoDataFrame with the polygons and metadata.
    """
    
    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    response = wfs.getfeature(
        typename=WFS_LAYER,
        bbox=tuple(bbox) + (crs,),
        outputFormat="json",
    )
    wb_gpd = gpd.read_file(response)
    return wb_gpd


def get_geohashes(bbox: tuple = None, crs: str = "EPSG:4326") -> [str]:
    """Gets all waterbody geohashes.
    
    Parameters
    ----------
    bbox : (xmin, ymin, xmax, ymax)
        Optional bounding box.
    crs : str
        Optional CRS for the bounding box.
    
    Returns
    -------
    [str]
        A list of geohashes.
    """
    
    wfs = WebFeatureService(url=WFS_ADDRESS, version="1.1.0")
    if bbox is not None:
        bbox = bbox + (crs,)
    response = wfs.getfeature(
        typename=WFS_LAYER,
        propertyname="UID",
        outputFormat="json",
        bbox=bbox,
    )
    wb_gpd = gpd.read_file(response)
    return list(wb_gpd["UID"])


def get_time_series(geohash: str = None, waterbody: pd.Series = None) -> pd.DataFrame:
    """Gets the time series for a waterbody. Specify either a GeoDataFrame row or a geohash.
    
    Parameters
    ----------
    geohash : str
        The geohash/UID for a waterbody in DE Africa Water Bodies.
    waterbody : pd.Series
        One row of a GeoDataFrame representing a waterbody.
    
    Returns
    -------
    pd.DataFrame
        A time series for the waterbody.
    """
    if waterbody is not None and geohash is not None:
        raise ValueError("One of waterbody and geohash must be None")
    if waterbody is None and geohash is None:
        raise ValueError("One of waterbody and geohash must be specified")

    if geohash is not None:
        wb = get_waterbody(geohash)
        url = wb.timeseries[0]
    else:
        url = waterbody.timeseries
    wb_timeseries = pd.read_csv(url)
    # Tidy up the dataframe.
    wb_timeseries.dropna(inplace=True)
    wb_timeseries.columns = ["date", "pc_wet", "px_wet", "area_wet_m2", "pc_dry", "px_dry", "area_dry_m2", "pc_invalid", "px_invalid", "area_invalid_m2"]
    wb_timeseries = wb_timeseries.set_index("date")
    wb_timeseries.index = pd.to_datetime(wb_timeseries.index)
    return wb_timeseries



"""
Wetlands insight tool widget, which can be used to run an interactive
version of the wetlands insight tool.
"""

# Import required packages

# Force GeoPandas to use Shapely instead of PyGEOS
# In a future release, GeoPandas will switch to using Shapely by default.
import os
os.environ['USE_PYGEOS'] = '0'

import datacube
import warnings
import seaborn as sns
import matplotlib.pyplot as plt
from datacube.utils.geometry import CRS
from ipyleaflet import (
    WMSLayer,
    basemaps,
    basemap_to_tiles,
    Map,
    DrawControl,
    WidgetControl,
    LayerGroup,
    LayersControl,
)
from traitlets import Unicode
from ipywidgets import (
    GridspecLayout,
    Button,
    Layout,
    HBox,
    VBox,
    HTML,
    Output,
)
import json
import geopandas as gpd
from io import BytesIO
from dask.diagnostics import ProgressBar

import deafrica_tools
from deafrica_tools.dask import create_local_dask_cluster
from deafrica_tools.wetlands import WIT_drill
import deafrica_tools.app.widgetconstructors as deawidgets


def make_box_layout():
     return Layout(
         #border='solid 1px black',
         margin='0px 10px 10px 0px',
         padding='5px 5px 5px 5px',
         width='100%',
         height='100%',
     )


def create_expanded_button(description, button_style):
    return Button(
        description=description,
        button_style=button_style,
        layout=Layout(width="auto", height="auto"),
    )


class wit_app(HBox):
    def __init__(self, lang=None):
        super().__init__()
        
        deafrica_tools.set_lang(lang)

        ##########################################################
        # INITIAL ATTRIBUTES #

        self.startdate = "2020-01-01"
        self.enddate = "2020-03-01"
        self.mingooddata = 0.0
        self.resamplingfreq = "1M"
        self.out_csv = "example_WIT.csv"
        self.out_plot = "example_WIT.png"
        self.product_list = [
            (_("None"), "none"),
            (_("ESRI World Imagery"), "esri_world_imagery"),
            (_("Sentinel-2 Geomedian"), "gm_s2_annual"),
            (_("Water Observations from Space"), "wofs_ls_summary_annual"),
            
        ]
        self.product = self.product_list[0][1]
        self.product_year = "2020-01-01"
        self.target = None
        self.action = None
        self.gdf_drawn = None
        
        ##########################################################
        # HEADER FOR APP #
        
        # Create the Header widget
        header_title_text = _("Wetlands Insight Tool")
        instruction_text = _("Select parameters and AOI")
        self.header = deawidgets.create_html(f"<h3>{header_title_text}</h3><p>{instruction_text}</p>")
        self.header.layout = make_box_layout()
        
        ##########################################################
        # HANDLER FUNCTION FOR DRAW CONTROL #
        
        # Define the action to take once something is drawn on the map
        def update_geojson(target, action, geo_json):

            self.action = action

            json_data = json.dumps(geo_json)
            binary_data = json_data.encode()
            io = BytesIO(binary_data)
            io.seek(0)

            gdf = gpd.read_file(io)
            gdf.crs = "EPSG:4326"
            self.gdf_drawn = gdf

            gdf_drawn_epsg6933 = gdf.copy().to_crs("EPSG:6933")
            m2_per_km2 = 10 ** 6
            area = gdf_drawn_epsg6933.area.values[0] / m2_per_km2
            polyarea_label = _('Total polygon area')
            polyarea_text = f"<p><b>{polyarea_label}</b>: {area:.2f} km<sup>2</sup></p>"

            if area <= 3000:
                confirmation_text = '<p style="color:#33cc33;">' + _('Area falls within recommended limit') + '</p>'
                self.header.value = header_title_text + polyarea_text + confirmation_text
            else:
                warning_text = '<p style="color:#ff5050;">' + _('Area is too large, please update your polygon') + '</p>'
                self.header.value = header_title_text + polyarea_text + warning_text

        ##########################################################
        # WIDGETS FOR APP OUTPUTS #

        self.dask_client = Output(layout=make_box_layout())
        self.progress_bar = Output(layout=make_box_layout())
        self.wit_plot = Output(layout=make_box_layout())
        self.progress_header = deawidgets.create_html("")

        ##########################################################
        # MAP WIDGET, DRAWING TOOLS, WMS LAYERS #

        # Create drawing tools
        desired_drawtools = ['rectangle', 'polygon']
        draw_control = deawidgets.create_drawcontrol(desired_drawtools)
        
        # Begin by displaying an empty layer group, and update the group with desired WMS on interaction.
        self.deafrica_layers = LayerGroup(layers=())
        self.deafrica_layers.name = _('Map Overlays')

        # Create map widget
        self.m = deawidgets.create_map()
        
        self.m.layout = make_box_layout()
        
        # Add tools to map widget
        self.m.add_control(draw_control)
        self.m.add_layer(self.deafrica_layers)
        
        # Store current basemap for future use
        self.basemap = self.m.basemap

        ##########################################################
        # WIDGETS FOR APP CONTROLS #

        # Create parameter widgets
        startdate_picker = deawidgets.create_datepicker()
        enddate_picker = deawidgets.create_datepicker()
        min_good_data = deawidgets.create_boundedfloattext(self.mingooddata, 0.0, 1.0, 0.05)
        resampling_freq = deawidgets.create_inputtext(self.resamplingfreq, self.resamplingfreq)
        output_csv = deawidgets.create_inputtext(self.out_csv, self.out_csv)
        output_plot = deawidgets.create_inputtext(self.out_plot, self.out_plot)
        deaoverlay_dropdown = deawidgets.create_dropdown(self.product_list, self.product_list[0][1])
        run_button = create_expanded_button(_("Run"), "info")

        ##########################################################
        # COLLECTION OF ALL APP CONTROLS #
        
        parameter_selection = VBox(
            [
                HTML("<b>" + _("Map Overlay:") + "</b>"),
                deaoverlay_dropdown,
                HTML("<b>" + _("Start Date:") + "</b>"),
                startdate_picker,
                HTML("<b>" + _("End Date:") + "</b>"),
                enddate_picker,
                HTML("<b>" + _("Minimum Good Data:") + "</b>"),
                min_good_data,
                HTML("<b>" + _("Resampling Frequency:") + "</b>"),
                resampling_freq,
                HTML("<b>" + _("Output CSV:") + "</b>"),
                output_csv,
                HTML("<b>" + _("Output Plot:") + "</b>"),
                output_plot,
            ]
        )
        parameter_selection.layout = make_box_layout()

        ##########################################################
        # SPECIFICATION OF APP LAYOUT #

        # Create the layout #[rowspan, colspan]
        grid = GridspecLayout(11, 10, height="1100px", width="auto")

        # Controls and Status
        grid[0, :] = self.header
        grid[1:6, 0:2] = parameter_selection
        grid[6, 0:2] = run_button
        
        # Dask and Progress info
        grid[1, 7:] = self.dask_client
        grid[2:7, 7:] = self.progress_bar

        # Map
        grid[1:7, 2:7] = self.m

        # Plot
        grid[7:, :] = self.wit_plot

        # Display using HBox children attribute
        self.children = [grid]

        ##########################################################
        # SPECIFICATION UPDATE FUNCTIONS FOR EACH WIDGET #

        # Run update functions whenever various widgets are changed.
        startdate_picker.observe(self.update_startdate, "value")
        enddate_picker.observe(self.update_enddate, "value")
        min_good_data.observe(self.update_mingooddata, "value")
        resampling_freq.observe(self.update_resamplingfreq, "value")
        output_csv.observe(self.update_outputcsv, "value")
        output_plot.observe(self.update_outputplot, "value")
        deaoverlay_dropdown.observe(self.update_deaoverlay, "value")
        run_button.on_click(self.run_app)
        draw_control.on_draw(update_geojson)

    ##############################################################
    # DEFINITION OF ALL UPDATE FUNCTIONS #

    # set the start date to the new edited date
    def update_startdate(self, change):
        self.startdate = change.new

    # set the end date to the new edited date
    def update_enddate(self, change):
        self.enddate = change.new

    # set the min good data
    def update_mingooddata(self, change):
        self.mingooddata = change.new

    # set the resampling frequency
    def update_resamplingfreq(self, change):
        self.resamplingfreq = change.new

    # set the output csv
    def update_outputcsv(self, change):
        self.out_csv = change.new

    # set the output plot
    def update_outputplot(self, change):
        self.out_plot = change.new

    # Update product
    def update_deaoverlay(self, change):

        self.product = change.new

        if self.product == "none":
            self.deafrica_layers.clear_layers()
        elif self.product == "esri_world_imagery":
            self.deafrica_layers.clear_layers()
            layer = basemap_to_tiles(basemaps.Esri.WorldImagery)
            self.deafrica_layers.add_layer(layer)
        else:
            self.deafrica_layers.clear_layers()
            layer = deawidgets.create_dea_wms_layer(self.product, self.product_year)
            self.deafrica_layers.add_layer(layer)

    def run_app(self, change):
        
        # Clear progress bar and output areas before running
        self.dask_client.clear_output()
        self.progress_bar.clear_output()
        self.wit_plot.clear_output()

        # Connect to datacube database
        dc = datacube.Datacube(app="wetland_app")

        # Configure local dask cluster
        with self.dask_client:
            client = create_local_dask_cluster(
                return_client=True, display_client=True
            )

        # Set any defaults
        TCW_threshold = -0.035
        dask_chunks = dict(x=1000, y=1000, time=1)
        
        #check resampling freq
        if self.resamplingfreq  == 'None':
            rsf = None
        else:
            rsf = self.resamplingfreq
        
        self.progress_header.value = f"<h3>"+_("Progress")+"</h3>"
            
        # run wetlands polygon drill
        with self.progress_bar:
#             with ProgressBar():
            warnings.filterwarnings("ignore")                
            try:
                df = WIT_drill(
                    gdf=self.gdf_drawn,
                    time=(self.startdate, self.enddate),
                    min_gooddata=self.mingooddata,
                    resample_frequency=rsf,
                    TCW_threshold=TCW_threshold,
                    export_csv=self.out_csv,
                    dask_chunks=dask_chunks,
                    verbose=False,
                    verbose_progress=True,
                )
                print(_("WIT complete"))
            except AttributeError:
                print(_("No polygon selected"))
        
        # close down the dask client
        client.shutdown()

        # save the csv
        if self.out_csv:
            df.to_csv(self.out_csv, index_label="Datetime")

        # ---Plotting------------------------------
        
        with self.wit_plot:

            fontsize = 17
            plt.rcParams.update({"font.size": fontsize})
            # set up color palette
            pal = [
                sns.xkcd_rgb["cobalt blue"],
                sns.xkcd_rgb["neon blue"],
                sns.xkcd_rgb["grass"],
                sns.xkcd_rgb["beige"],
                sns.xkcd_rgb["brown"],
            ]

            # make a stacked area plot
            plt.close("all")

            fig, ax = plt.subplots(constrained_layout=True, figsize=(20, 6))

            ax.stackplot(
                df.index,
                df.wofs_area_percent,
                df.wet_percent,
                df.green_veg_percent,
                df.dry_veg_percent,
                df.bare_soil_percent,
                labels=[
                    _("open water"),
                    _("wet"),
                    _("green veg"),
                    _("dry veg"),
                    _("bare soil"),
                ],
                colors=pal,
                alpha=0.6,
            )

            # set axis limits to the min and max
            ax.set_ylim(0, 100)
            ax.set_xlim(df.index[0], df.index[-1])
            ax.tick_params(axis="x", labelsize=fontsize)

            # add a legend and a tight plot box
            ax.legend(loc="lower left", framealpha=0.6)
            ax.set_title(_("Percentage Fractional Cover, Wetness, and Water"))
            # plt.tight_layout()
            plt.show()

            if self.out_plot:
                # save the figure
                fig.savefig(f"{self.out_plot}")



"""
Functions for loading and handling Digital Earth Africa data.
"""

# Import required packages
import os
from osgeo import gdal
import requests
import zipfile
import warnings
import numpy as np
import xarray as xr
import pandas as pd
import datetime
import pytz

from collections import Counter
from datacube.utils import masking
from odc.algo import mask_cleanup
from copy import deepcopy
import odc.algo

from skimage.morphology import binary_erosion,binary_dilation,disk
from scipy.ndimage import binary_dilation
from scipy.ndimage.filters import uniform_filter
from scipy.ndimage.measurements import variance
from datetime import datetime
from dateutil import parser
from deafrica_tools.bandindices import calculate_indices

def _dc_query_only(**kw):
    """
    Remove load-only parameters, the rest
    can be passed to Query

    Returns
    =======

    dict of query parameters
    """

    def _impl(
        measurements=None,
        output_crs=None,
        resolution=None,
        resampling=None,
        skip_broken_datasets=None,
        dask_chunks=None,
        fuse_func=None,
        align=None,
        datasets=None,
        progress_cbk=None,
        group_by=None,
        **query,
    ):
        return query

    return _impl(**kw)


def _common_bands(dc, products):
    """
    Takes a list of products and returns a list of measurements/bands
    that are present in all products
    Returns
    -------
    List of band names
    """
    common = None
    bands = None

    for p in products:
        p = dc.index.products.get_by_name(p)
        if common is None:
            common = set(p.measurements)
            bands = list(p.measurements)
        else:
            common = common.intersection(set(p.measurements))
    return [band for band in bands if band in common]


def load_ard(
    dc,
    products=None,
    min_gooddata=0.0,
    categories_to_mask_ls=dict(
        cloud="high_confidence", cloud_shadow="high_confidence"
    ),
    categories_to_mask_s2=[
        "cloud high probability",
        "cloud medium probability",
        "thin cirrus",
        "cloud shadows",
        "saturated or defective",
    ],
    categories_to_mask_s1=["invalid data"],
    mask_filters=None,
    mask_pixel_quality=True,
    ls7_slc_off=True,
    predicate=None,
    dtype="auto",
    verbose=True,
    **kwargs,
):
    """
    Loads analysis ready data.

    Loads and combines Landsat USGS Collections 2, Sentinel-2, and Sentinel-1 for
    multiple sensors (i.e. ls5t, ls7e, ls8c and ls9 for Landsat; s2a and s2b for Sentinel-2),
    optionally applies pixel quality masks, and drops time steps that
    contain greater than a minimum proportion of good quality (e.g. non-
    cloudy or shadowed) pixels.

    The function supports loading the following DE Africa products:

    Landsat:
        * ls5_sr ('sr' denotes surface reflectance)
        * ls7_sr
        * ls8_sr
        * ls9_sr
        * ls5_st ('st' denotes surface temperature)
        * ls7_st
        * ls8_st
        * ls9_st

    Sentinel-2:
        * s2_l2a

    Sentinel-1:
        * s1_rtc

    Last modified: Feb 2021

    Parameters
    ----------
    dc : datacube Datacube object
        The Datacube to connect to, i.e. `dc = datacube.Datacube()`.
        This allows you to also use development datacubes if required.
    products : list
        A list of product names to load data from. For example:

        * Landsat C2: ``['ls5_sr', 'ls7_sr', 'ls8_sr', 'ls9_sr']``
        * Sentinel-2: ``['s2_l2a']``
        * Sentinel-1: ``['s1_rtc']``

    min_gooddata : float, optional
        An optional float giving the minimum percentage of good quality
        pixels required for a satellite observation to be loaded.
        Defaults to 0.0 which will return all observations regardless of
        pixel quality (set to e.g. 0.99 to return only observations with
        more than 99% good quality pixels).
    categories_to_mask_ls : dict, optional
        An optional dictionary that is used to identify poor quality pixels
        for masking. This mask is used for both masking out low
        quality pixels (e.g. cloud or shadow), and for dropping
        observations entirely based on the `min_gooddata` calculation.
    categories_to_mask_s2 : list, optional
        An optional list of Sentinel-2 Scene Classification Layer (SCL) names
        that identify poor quality pixels for masking.
    categories_to_mask_s1 : list, optional
        An optional list of Sentinel-1 mask names that identify poor
        quality pixels for masking.
    mask_filters : iterable of tuples, optional
        Iterable tuples of morphological operations - ("<operation>", <radius>)
        to apply on mask, where:

        operation: string, can be one of these morphological operations:
            * ``'closing'``  = remove small holes in cloud - morphological closing
            * ``'opening'``  = shrinks away small areas of the mask
            * ``'dilation'`` = adds padding to the mask
            * ``'erosion'``  = shrinks bright regions and enlarges dark regions

        radius: int
        e.g. ``mask_filters=[('erosion', 5),("opening", 2),("dilation", 2)]``
    mask_pixel_quality : bool, optional
        An optional boolean indicating whether to apply the poor data
        mask to all observations that were not filtered out for having
        less good quality pixels than ``min_gooddata``. E.g. if
        ``min_gooddata=0.99``, the filtered observations may still contain
        up to 1% poor quality pixels. The default of ``False`` simply
        returns the resulting observations without masking out these
        pixels; ``True`` masks them and sets them to NaN using the poor data
        mask. This will convert numeric values to floating point values
        which can cause memory issues, set to False to prevent this.
    ls7_slc_off : bool, optional
        An optional boolean indicating whether to include data from
        after the Landsat 7 SLC failure (i.e. SLC-off). Defaults to
        ``True``, which keeps all Landsat 7 observations > May 31 2003.
    predicate : function, optional
        An optional function that can be passed in to restrict the
        datasets that are loaded by the function. A filter function
        should take a `datacube.model.Dataset` object as an input (i.e.
        as returned from `dc.find_datasets`), and return a boolean.
        For example, a filter function could be used to return True on
        only datasets acquired in January:
        ``dataset.time.begin.month == 1``
    dtype : string, optional
        An optional parameter that controls the data type/dtype that
        layers are coerced to after loading. Valid values: ''`native`'',
        ``'auto'``, ``'float{16|32|64}'``.
        When ``'auto'`` is used, the data will be
        converted to ``'float32'`` if masking is used, otherwise data will
        be returned in the native data type of the data. Be aware that
        if data is loaded in its native dtype, nodata and masked
        pixels will be returned with the data's native nodata value
        (typically ``-999``), not ``NaN``.
        NOTE: If loading Landsat, the data is automatically rescaled so
        'native' dtype will return a value error.
    verbose : bool, optional
        If True, print progress statements during loading
    **kwargs : dict, optional
        A set of keyword arguments to ``dc.load`` that define the
        spatiotemporal query used to extract data. This typically
        includes ``measurements``, ``x`, ``y``, ``time``, ``resolution``,
        ``resampling``, ``group_by`` and ``crs``. Keyword arguments can
        either be listed directly in the ``load_ard`` call like any
        other parameter (e.g. ``measurements=['red']``), or by
        passing in a query kwarg dictionary (e.g. ``**query``). For a
        list of possible options, see the ``dc.load`` documentation:
        https://datacube-core.readthedocs.io/en/latest/dev/api/generate/datacube.Datacube.load.html

    Returns
    -------
    combined_ds : xarray Dataset
        An xarray dataset containing only satellite observations that
        contains greater than `min_gooddata` proportion of good quality
        pixels.

    """

    #########
    # Setup #
    #########
    # prevent function altering original query object
    kwargs = deepcopy(kwargs)

    # We deal with `dask_chunks` separately
    dask_chunks = kwargs.pop("dask_chunks", None)
    requested_measurements = kwargs.pop("measurements", None)

    # Warn user if they combine lazy load with min_gooddata
    if verbose:
        if (min_gooddata > 0.0) and dask_chunks is not None:
            warnings.warn(
                "Setting 'min_gooddata' percentage to > 0.0 "
                "will cause dask arrays to compute when "
                "loading pixel-quality data to calculate "
                "'good pixel' percentage. This can "
                "slow the return of your dataset."
            )

    # Verify that products were provided and determine if Sentinel-2
    # or Landsat data is being loaded
    if not products:
        raise ValueError(
            "Please provide a list of product names to load data from. "
            "Valid options are: Landsat C2 SR: ['ls5_sr', 'ls7_sr', 'ls8_sr', 'ls9_sr'], or "
            "Landsat C2 ST: ['ls5_st', 'ls7_st', 'ls8_st', 'ls9_st'], or "
            "Sentinel-2: ['s2_l2a'], or"
            "Sentinel-1: ['s1_rtc'], or"
        )
    
    # convert products to list if user passed as a string
    if type(products) == str:
        products=[products]
    
    if all(["ls" in product for product in products]):
        product_type = "ls"
    elif all(["s2" in product for product in products]):
        product_type = "s2"
    elif all(["s1" in product for product in products]):
        product_type = "s1"

    # check if the landsat product is surface temperature
    st = False
    if (product_type == "ls") & (all(["st" in product for product in products])):
        st = True

    # Check some parameters before proceeding
    if (product_type == "ls") & (dtype == "native"):
        raise ValueError(
            "Cannot load Landsat bands in native dtype "
            "as values require rescaling which converts dtype to float"
        )

    if product_type == "ls":
        if any(k in categories_to_mask_ls for k in ("cirrus", "cirrus_confidence")):
            raise ValueError(
                "'cirrus' categories for the pixel quality mask"
                " are not supported by load_ard"
            )

    # If `measurements` are specified but do not include pixel quality bands,
    #  add these to `measurements` according to collection
    if product_type == "ls":
        if verbose:
            print("Using pixel quality parameters for USGS Collection 2")
        fmask_band = "pixel_quality"

    elif product_type == "s2":
        if verbose:
            print("Using pixel quality parameters for Sentinel 2")
        fmask_band = "SCL"

    elif product_type == "s1":
        if verbose:
            print("Using pixel quality parameters for Sentinel 1")
        fmask_band = "mask"

    measurements = requested_measurements.copy() if requested_measurements else None

    # define a list of acceptable aliases to load landsat. We can't rely on 'common'
    # measurements as native band names have the same name for different measurements.
    ls_aliases = ["pixel_quality", "radiometric_saturation"]
    if st:
        ls_aliases = [
            "surface_temperature",
            "surface_temperature_quality",
            "atmospheric_transmittance",
            "thermal_radiance",
            "emissivity",
            "emissivity_stddev",
            "cloud_distance",
            "upwell_radiance",
            "downwell_radiance",
        ] + ls_aliases
    else:
        ls_aliases = ["red", "green", "blue", "nir", "swir_1", "swir_2"] + ls_aliases

    if measurements is not None:
        if product_type == "ls":

            # check we aren't loading aerosol bands from LS8
            aerosol_bands = [
                "aerosol_qa",
                "qa_aerosol",
                "atmos_opacity",
                "coastal_aerosol",
                "SR_QA_AEROSOL",
            ]
            if any(b in aerosol_bands for b in measurements):
                raise ValueError(
                    "load_ard doesn't support loading aerosol or "
                    "atmospeheric opacity related bands "
                    "for Landsat, instead use dc.load()"
                )

            # check measurements are in acceptable aliases list for landsat
            if set(measurements).issubset(ls_aliases):
                pass
            else:
                raise ValueError(
                    "load_ard does not support all band aliases for Landsat, "
                    "use only the following band names to load Landsat data: "
                    + str(ls_aliases)
                )

    # Deal with "load all" case: pick a set of bands common across
    # all products
    if measurements is None:
        if product_type == "ls":
            measurements = ls_aliases
        else:
            measurements = _common_bands(dc, products)

    # If `measurements` are specified but do not include pq, add.
    if measurements:
        if fmask_band not in measurements:
            measurements.append(fmask_band)

    # Get list of data and mask bands so that we can later exclude
    # mask bands from being masked themselves (also handle the case of rad_sat)
    data_bands = [
        band
        for band in measurements
        if band not in (fmask_band, "radiometric_saturation")
    ]
    mask_bands = [band for band in measurements if band not in data_bands]

    #################
    # Find datasets #
    #################

    # Pull out query params only to pass to dc.find_datasets
    query = _dc_query_only(**kwargs)

    # Extract datasets for each product using subset of dcload_kwargs
    dataset_list = []

    # Get list of datasets for each product
    if verbose:
        print("Finding datasets")
    for product in products:

        # Obtain list of datasets for product
        if verbose:
            print(f"    {product}")

        if product_type == "ls":
            # handle LS seperately to S2/S1 due to collection_category
            # force the user to load Tier 1
            datasets = dc.find_datasets(
                product=product, collection_category='T1', **query
            )
        else:
            datasets = dc.find_datasets(product=product, **query)

        # Remove Landsat 7 SLC-off observations if ls7_slc_off=False
        if not ls7_slc_off and product in ["ls7_sr"]:
            if verbose:
                print("    Ignoring SLC-off observations for ls7")
            datasets = [
                i
                for i in datasets
                if i.time.begin < datetime(2003, 5, 31, tzinfo=pytz.UTC)
            ]

        # Add any returned datasets to list
        dataset_list.extend(datasets)

    # Raise exception if no datasets are returned
    if len(dataset_list) == 0:
        raise ValueError(
            "No data available for query: ensure that "
            "the products specified have data for the "
            "time and location requested"
        )

    # If predicate is specified, use this function to filter the list
    # of datasets prior to load (this now redundant as dc.load now supports
    # a predicate filter)
    if predicate:
        if verbose:
            print(f"Filtering datasets using filter function")
        dataset_list = [ds for ds in dataset_list if predicate(ds)]

    # Raise exception if filtering removes all datasets
    if len(dataset_list) == 0:
        raise ValueError("No data available after filtering with " "filter function")

    #############
    # Load data #
    #############

    # Note we always load using dask here so that
    # we can lazy load data before filtering by good data
    ds = dc.load(
        datasets=dataset_list,
        measurements=measurements,
        dask_chunks={} if dask_chunks is None else dask_chunks,
        **kwargs,
    )

    ####################
    # Filter good data #
    ####################

    # need to distinguish between products due to different
    # pq band properties

    # collection 2 USGS
    if product_type == "ls":
        mask, _ = masking.create_mask_value(
            ds[fmask_band].attrs["flags_definition"], **categories_to_mask_ls
        )
        
        pq_mask = (ds[fmask_band] & mask) != 0
        
        # only run if data bands are present 
        if len(data_bands) > 0: 
            
        # identify pixels that will become negative after rescaling (but not 0 values)
            invalid = (
                    ((ds[data_bands] < (-1.0 * -0.2 / 0.0000275)) & (ds[data_bands] > 0))
                    .to_array(dim="band")
                    .any(dim="band")
                    )

        #merge masks
        pq_mask = np.logical_or(pq_mask, pq_mask)

    # sentinel 2
    if product_type == "s2":
        pq_mask = odc.algo.enum_to_bool(mask=ds[fmask_band],
                                        categories=categories_to_mask_s2)
        
    # sentinel 1
    if product_type == "s1":
        pq_mask = odc.algo.enum_to_bool(mask=ds[fmask_band],
                                        categories=categories_to_mask_s1)

    # The good data percentage calculation has to load in all `fmask`
    # data, which can be slow. If the user has chosen no filtering
    # by using the default `min_gooddata = 0`, we can skip this step
    # completely to save processing time
    if min_gooddata > 0.0:

        # Compute good data for each observation as % of total pixels.
        # Inveerting the pq_mask for this because cloud=True in pq_mask
        # and we want to sum good pixels
        if verbose:
            print("Counting good quality pixels for each time step")
        data_perc = (~pq_mask).sum(axis=[1, 2], dtype="int32") / (
            pq_mask.shape[1] * pq_mask.shape[2]
        )

        keep = (data_perc >= min_gooddata).persist()

        # Filter by `min_gooddata` to drop low quality observations
        total_obs = len(ds.time)
        ds = ds.sel(time=keep)
        pq_mask = pq_mask.sel(time=keep)

        if verbose:
            print(
                f"Filtering to {len(ds.time)} out of {total_obs} "
                f"time steps with at least {min_gooddata:.1%} "
                f"good quality pixels"
            )

    # morpholigcal filtering on cloud masks
    if (mask_filters is not None) & (mask_pixel_quality):
        if verbose:
            print(f"Applying morphological filters to pq mask {mask_filters}")
        pq_mask = mask_cleanup(pq_mask, mask_filters=mask_filters)

    ###############
    # Apply masks #
    ###############

    # Generate good quality data mask
    mask = None
    if mask_pixel_quality:
        if verbose:
            print("Applying pixel quality/cloud mask")
        mask = pq_mask

    # Split into data/masks bands, as conversion to float and masking
    # should only be applied to data bands
    ds_data = ds[data_bands]
    ds_masks = ds[mask_bands]
    
    # Remove sentinel-2 pixels valued 1 (scene edges, terrain shadow)
    if product_type == "s2":
        valid_data_mask = (ds_data > 1).to_array(dim="band").all(dim="band")
        ds_data =  odc.algo.keep_good_only(ds_data, where=valid_data_mask)
        
    # Mask data if either of the above masks were generated
    if mask is not None:
        ds_data = odc.algo.erase_bad(ds_data, where=mask)
    
    # Automatically set dtype to either native or float32 depending
    # on whether masking was requested
    if dtype == "auto":
        dtype = "native" if mask is None else "float32"

    # Set nodata values using odc.algo tools to reduce peak memory
    # use when converting data dtype
    if dtype != "native":
        ds_data = odc.algo.to_float(ds_data, dtype=dtype)

    # Put data and mask bands back together
    attrs = ds.attrs
    ds = xr.merge([ds_data, ds_masks])
    ds.attrs.update(attrs)

    ###############
    # Return data #
    ###############

    # Drop bands not originally requested by user
    if requested_measurements:
        ds = ds[requested_measurements]

    # Apply the scale and offset factors to Collection 2 Landsat. We need
    # different factors for different bands. Also handle the case where
    # masking_pixel_quaity = False, in which case the dtype is still
    # in int, so we convert it to float
    if product_type == "ls":
        if verbose:
            print("Re-scaling Landsat C2 data")

        sr_bands = ["red", "green", "blue", "nir", "swir_1", "swir_2"]
        radiance_bands = ["thermal_radiance", "upwell_radiance", "downwell_radiance"]
        trans_emiss = ["atmospheric_transmittance", "emissivity", "emissivity_stddev"]
        qa = ["pixel_quality", "radiometric_saturation"]

        if mask_pixel_quality == False:
            # set nodata to NaNs before rescaling
            # in the case where masking hasn't already done this
            for band in ds.data_vars:
                if band not in qa:
                    ds[band] = odc.algo.to_f32(ds[band])

        for band in ds.data_vars:
            if band == "cloud_distance":
                ds[band] = 0.01 * ds[band]

            if band == "surface_temperature_quality":
                ds[band] = 0.01 * ds[band]

            if band in radiance_bands:
                ds[band] = 0.001 * ds[band]

            if band in trans_emiss:
                ds[band] = 0.0001 * ds[band]

            if band in sr_bands:
                ds[band] = 2.75e-5 * ds[band] - 0.2

            if band == "surface_temperature":
                ds[band] = ds[band] * 0.00341802 + 149.0

        # add back attrs that are lost during scaling calcs
        for band in ds.data_vars:
            ds[band].attrs.update(attrs)

    # If user supplied dask_chunks, return data as a dask array without
    # actually loading it in
    if dask_chunks is not None:
        if verbose:
            print(f"Returning {len(ds.time)} time steps as a dask array")
        return ds
    else:
        if verbose:
            print(f"Loading {len(ds.time)} time steps")
        return ds.compute()


def array_to_geotiff(
    fname, data, geo_transform, projection, nodata_val=0, dtype=gdal.GDT_Float32
):
    """
    Create a single band GeoTIFF file with data from an array.

    Because this works with simple arrays rather than xarray datasets
    from DEA, it requires geotransform info (`(upleft_x, x_size,
    x_rotation, upleft_y, y_rotation, y_size)`) and projection data
    (in "WKT" format) for the output raster. These are typically
    obtained from an existing raster using the following GDAL calls:

    >>> from osgeo import gdal
    >>> gdal_dataset = gdal.Open(raster_path)
    >>> geotrans = gdal_dataset.GetGeoTransform()
    >>> prj = gdal_dataset.GetProjection()

    or alternatively, directly from an xarray dataset:

    >>> geotrans = xarraydataset.geobox.transform.to_gdal()
    >>> prj = xarraydataset.geobox.crs.wkt


    Parameters
    ----------
    fname : str
        Output geotiff file path including extension
    data : numpy array
        Input array to export as a geotiff
    geo_transform : tuple
        Geotransform for output raster; e.g. `(upleft_x, x_size,
        x_rotation, upleft_y, y_rotation, y_size)`
    projection : str
        Projection for output raster (in "WKT" format)
    nodata_val : int, optional
        Value to convert to nodata in the output raster; default 0
    dtype : gdal dtype object, optional
        Optionally set the dtype of the output raster; can be
        useful when exporting an array of float or integer values.
        Defaults to `gdal.GDT_Float32`

    """

    # Set up driver
    driver = gdal.GetDriverByName("GTiff")

    # Create raster of given size and projection
    rows, cols = data.shape
    dataset = driver.Create(fname, cols, rows, 1, dtype)
    dataset.SetGeoTransform(geo_transform)
    dataset.SetProjection(projection)

    # Write data to array and set nodata values
    band = dataset.GetRasterBand(1)
    band.WriteArray(data)
    band.SetNoDataValue(nodata_val)

    # Close file
    dataset = None


def mostcommon_crs(dc, product, query):
    """
    Takes a given query and returns the most common CRS for observations
    returned for that spatial extent. This can be useful when your study
    area lies on the boundary of two UTM zones, forcing you to decide
    which CRS to use for your `output_crs` in `dc.load`.

    Parameters
    ----------
    dc : datacube Datacube object
        The Datacube to connect to, i.e. `dc = datacube.Datacube()`.
        This allows you to also use development datacubes if required.
    product : str
        A product name to load CRSs from
    query : dict
        A datacube query including x, y and time range to assess for the
        most common CRS

    Returns
    -------
    str
        A EPSG string giving the most common CRS from all datasets returned
        by the query above

    """

    # remove dask_chunks & align to prevent func failing
    # prevent function altering dictionary kwargs
    query = deepcopy(query)
    if "dask_chunks" in query:
        query.pop("dask_chunks", None)

    if "align" in query:
        query.pop("align", None)

    # List of matching products
    matching_datasets = dc.find_datasets(product=product, **query)

    # Extract all CRSs
    crs_list = [str(i.crs) for i in matching_datasets]

    # Identify most common CRS
    crs_counts = Counter(crs_list)
    crs_mostcommon = crs_counts.most_common(1)[0][0]

    # Warn user if multiple CRSs are encountered
    if len(crs_counts.keys()) > 1:

        warnings.warn(
            f"Multiple UTM zones {list(crs_counts.keys())} "
            f"were returned for this query. Defaulting to "
            f"the most common zone: {crs_mostcommon}",
            UserWarning,
        )

    return crs_mostcommon


def download_unzip(url, output_dir=None, remove_zip=True):
    """
    Downloads and unzips a .zip file from an external URL to a local
    directory.

    Parameters
    ----------
    url : str
        A string giving a URL path to the zip file you wish to download
        and unzip
    output_dir : str, optional
        An optional string giving the directory to unzip files into.
        Defaults to None, which will unzip files in the current working
        directory
    remove_zip : bool, optional
        An optional boolean indicating whether to remove the downloaded
        .zip file after files are unzipped. Defaults to True, which will
        delete the .zip file.

    """

    # Get basename for zip file
    zip_name = os.path.basename(url)

    # Raise exception if the file is not of type .zip
    if not zip_name.endswith(".zip"):
        raise ValueError(
            f"The URL provided does not point to a .zip "
            f"file (e.g. {zip_name}). Please specify a "
            f"URL path to a valid .zip file"
        )

    # Download zip file
    print(f"Downloading {zip_name}")
    r = requests.get(url)
    with open(zip_name, "wb") as f:
        f.write(r.content)

    # Extract into output_dir
    with zipfile.ZipFile(zip_name, "r") as zip_ref:
        zip_ref.extractall(output_dir)
        print(
            f"Unzipping output files to: "
            f"{output_dir if output_dir else os.getcwd()}"
        )

    # Optionally cleanup
    if remove_zip:
        os.remove(zip_name)


def wofs_fuser(dest, src):
    """
    Fuse two WOfS water measurements represented as `ndarray` objects.

    Note: this is a copy of the function located here:
    https://github.com/GeoscienceAustralia/digitalearthau/blob/develop/digitalearthau/utils.py
    """
    empty = (dest & 1).astype(bool)
    both = ~empty & ~((src & 1).astype(bool))
    dest[empty] = src[empty]
    dest[both] |= src[both]


def dilate(array, dilation=10, invert=True):
    """
    Dilate a binary array by a specified nummber of pixels using a
    disk-like radial dilation.

    By default, invalid (e.g. False or 0) values are dilated. This is
    suitable for applications such as cloud masking (e.g. creating a
    buffer around cloudy or shadowed pixels). This functionality can
    be reversed by specifying `invert=False`.

    Parameters
    ----------
    array : array
        The binary array to dilate.
    dilation : int, optional
        An optional integer specifying the number of pixels to dilate
        by. Defaults to 10, which will dilate `array` by 10 pixels.
    invert : bool, optional
        An optional boolean specifying whether to invert the binary
        array prior to dilation. The default is True, which dilates the
        invalid values in the array (e.g. False or 0 values).

    Returns
    -------
    array
        An array of the same shape as `array`, with valid data pixels
        dilated by the number of pixels specified by `dilation`.
    """

    y, x = np.ogrid[
        -dilation : (dilation + 1),
        -dilation : (dilation + 1),
    ]

    # disk-like radial dilation
    kernel = (x * x) + (y * y) <= (dilation + 0.5) ** 2

    # If invert=True, invert True values to False etc
    if invert:
        array = ~array

    return ~binary_dilation(
        array.astype(bool), structure=kernel.reshape((1,) + kernel.shape)
    )


def _select_along_axis(values, idx, axis):
    other_ind = np.ix_(*[np.arange(s) for s in idx.shape])
    sl = other_ind[:axis] + (idx,) + other_ind[axis:]
    return values[sl]


def first(array: xr.DataArray, dim: str, index_name: str = None) -> xr.DataArray:
    """
    Finds the first occuring non-null value along the given dimension.

    Parameters
    ----------
    array : xr.DataArray
         The array to search.
    dim : str
        The name of the dimension to reduce by finding the first non-null value.

    Returns
    -------
    reduced : xr.DataArray
        An array of the first non-null values.
        The `dim` dimension will be removed, and replaced with a coord of the
        same name, containing the value of that dimension where the last value
        was found.
    """
    axis = array.get_axis_num(dim)
    idx_first = np.argmax(~pd.isnull(array), axis=axis)
    reduced = array.reduce(_select_along_axis, idx=idx_first, axis=axis)
    reduced[dim] = array[dim].isel({dim: xr.DataArray(idx_first, dims=reduced.dims)})
    if index_name is not None:
        reduced[index_name] = xr.DataArray(idx_first, dims=reduced.dims)
    return reduced


def last(array: xr.DataArray, dim: str, index_name: str = None) -> xr.DataArray:
    """
    Finds the last occuring non-null value along the given dimension.

    Parameters
    ----------
    array : xr.DataArray
         The array to search.
    dim : str
        The name of the dimension to reduce by finding the last non-null value.
    index_name : str, optional
        If given, the name of a coordinate to be added containing the index
        of where on the dimension the nearest value was found.

    Returns
    -------
    reduced : xr.DataArray
        An array of the last non-null values.
        The `dim` dimension will be removed, and replaced with a coord of the
        same name, containing the value of that dimension where the last value
        was found.
    """
    axis = array.get_axis_num(dim)
    rev = (slice(None),) * axis + (slice(None, None, -1),)
    idx_last = -1 - np.argmax(~pd.isnull(array)[rev], axis=axis)
    reduced = array.reduce(_select_along_axis, idx=idx_last, axis=axis)
    reduced[dim] = array[dim].isel({dim: xr.DataArray(idx_last, dims=reduced.dims)})
    if index_name is not None:
        reduced[index_name] = xr.DataArray(idx_last, dims=reduced.dims)
    return reduced


def nearest(
    array: xr.DataArray, dim: str, target, index_name: str = None
) -> xr.DataArray:
    """
    Finds the nearest values to a target label along the given dimension, for
    all other dimensions.

    E.g. For a DataArray with dimensions ('time', 'x', 'y')

        nearest_array = nearest(array, 'time', '2017-03-12')

    will return an array with the dimensions ('x', 'y'), with non-null values
    found closest for each (x, y) pixel to that location along the time
    dimension.

    The returned array will include the 'time' coordinate for each x,y pixel
    that the nearest value was found.

    Parameters
    ----------
    array : xr.DataArray
         The array to search.
    dim : str
        The name of the dimension to look for the target label.
    target : same type as array[dim]
        The value to look up along the given dimension.
    index_name : str, optional
        If given, the name of a coordinate to be added containing the index
        of where on the dimension the nearest value was found.

    Returns
    -------
    nearest_array : xr.DataArray
        An array of the nearest non-null values to the target label.
        The `dim` dimension will be removed, and replaced with a coord of the
        same name, containing the value of that dimension closest to the
        given target label.
    """
    before_target = slice(None, target)
    after_target = slice(target, None)

    da_before = array.sel({dim: before_target})
    da_after = array.sel({dim: after_target})

    da_before = last(da_before, dim, index_name) if da_before[dim].shape[0] else None
    da_after = first(da_after, dim, index_name) if da_after[dim].shape[0] else None

    if da_before is None and da_after is not None:
        return da_after
    if da_after is None and da_before is not None:
        return da_before

    target = array[dim].dtype.type(target)
    is_before_closer = abs(target - da_before[dim]) < abs(target - da_after[dim])
    nearest_array = xr.where(is_before_closer, da_before, da_after)
    nearest_array[dim] = xr.where(is_before_closer, da_before[dim], da_after[dim])
    if index_name is not None:
        nearest_array[index_name] = xr.where(
            is_before_closer, da_before[index_name], da_after[index_name]
        )
    return nearest_array

def parallel_apply(ds, dim, func, *args):
    """
    Applies a custom function in parallel along the dimension of an
    xarray.Dataset or xarray.DataArray.

    The function can be any function that can be applied to an
    individual xarray.Dataset or xarray.DataArray (e.g. data for a
    single timestep). The function should also return data in
    xarray.Dataset or xarray.DataArray format.

    This function is useful as a simple method for parallising code
    that cannot easily be parallised using Dask.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
        xarray data with a dimension `dim` to apply the custom function
        along.
    dim : string
        The dimension along which the custom function will be applied.
    func : function
        The function that will be applied in parallel to each array
        along dimension `dim`. The first argument passed to this
        function should be the array along `dim`.
    *args :
        Any number of arguments that will be passed to `func`.

    Returns
    -------
    xarray.Dataset
        A concatenated dataset containing an output for each array
        along the input `dim` dimension.
    """

    from concurrent.futures import ProcessPoolExecutor
    from tqdm import tqdm
    from itertools import repeat

    with ProcessPoolExecutor() as executor:

        # Apply func in parallel
        groups = [group for (i, group) in ds.groupby(dim)]
        to_iterate = (groups, *(repeat(i, len(groups)) for i in args))
        out_list = list(tqdm(executor.map(func, *to_iterate), total=len(groups)))

    # Combine to match the original dataset
    return xr.concat(out_list, dim=ds[dim])


def pan_sharpen_brovey(band_1, band_2, band_3, pan_band):
    """
    Brovey pan sharpening on surface reflectance input using numexpr
    and return three xarrays.
    Parameters
    ----------
    band_1, band_2, band_3 : xarray.DataArray or numpy.array
        Three input multispectral bands, either as xarray.DataArrays or
        numpy.arrays. These bands should have already been resampled to
        the spatial resolution of the panchromatic band.
    pan_band : xarray.DataArray or numpy.array
        A panchromatic band corresponding to the above multispectral
        bands that will be used to pan-sharpen the data.
    Returns
    -------
    band_1_sharpen, band_2_sharpen, band_3_sharpen : numpy.arrays
        Three numpy arrays equivelent to `band_1`, `band_2` and `band_3`
        pan-sharpened to the spatial resolution of `pan_band`.
    """
    # Calculate total
    exp = 'band_1 + band_2 + band_3'
    total = numexpr.evaluate(exp)

    # Perform Brovey Transform in form of: band/total*panchromatic
    exp = 'a/b*c'
    band_1_sharpen = numexpr.evaluate(exp, local_dict={'a': band_1,
                                                       'b': total,
                                                       'c': pan_band})
    band_2_sharpen = numexpr.evaluate(exp, local_dict={'a': band_2,
                                                       'b': total,
                                                       'c': pan_band})
    band_3_sharpen = numexpr.evaluate(exp, local_dict={'a': band_3,
                                                       'b': total,
                                                       'c': pan_band})

    return band_1_sharpen, band_2_sharpen, band_3_sharpen

def load_s1_by_orbits(dc,query):
    '''
    Function to query and load ascending and descending Sentinel-1 data 
    and add a variable to denote acquisition orbits
    
    Parameters:
    dc: connected datacube
    query: a query dictionary to define spatial extent, measurements, time range and spatial resolution
    
    Returns:
    Queried dataset with variable 'is_ascending' added to denote orbit path
    
    '''
    # load ascending data
    print('\nQuerying and loading Sentinel-1 ascending data...')
    ds_s1_ascending=load_ard(dc=dc,products=['s1_rtc'],resampling='bilinear',
                             dtype='native',sat_orbit_state='ascending',**query)
    # add an variable denoting data source
    ds_s1_ascending['is_ascending']=xr.DataArray(np.ones(len(ds_s1_ascending.time)),
                                                 dims=('time'),coords={'time': ds_s1_ascending.time})
    
    # load descending data
    print('\nQuerying and loading Sentinel-1 descending data...')
    ds_s1_descending=load_ard(dc=dc,products=['s1_rtc'],resampling='bilinear',
                              dtype='native',sat_orbit_state='descending',**query)
    # add an variable denoting data source
    ds_s1_descending['is_ascending']=xr.DataArray(np.zeros(len(ds_s1_descending.time)),
                                                  dims=('time'),coords={'time': ds_s1_descending.time})
    
    # merge datasets together
    ds_s1=xr.concat([ds_s1_ascending,ds_s1_descending],dim='time').sortby('time')
    
    return ds_s1

def filter_obs_by_orbit(ds_s1):
    '''
    Function to impliment per-pixel filtering of Sentinel-1 observations 
    to keep only observations from the orbit (ascending/descending) with higher frequency over time. 
    
    Each of the Sentinel-1 observations was acquired from either a descending or ascending orbit, 
    which has impacts on the local incidence angle and backscattering value. 
    Here we do the filtering to minimise the effects of inconsistent looking angle and obit direction for each individual pixel.

    Parameters:
    ds_s1: xarray.Dataset
        Time-series observations of Sentinel-1 data, 
        with two required variables: 'is_ascending' denoting orbit path and 'mask' to identify acquisition exent
    
    Returns:
    ds_s1_filtered: xarray.Dataset
        Filtered dataset
    '''

    print('\nFiltering Sentinel-1 product by orbit...')
    cnt_ascending=((ds_s1["is_ascending"]==1)&(ds_s1['mask']!=0)).sum(dim='time')
    cnt_descending=((ds_s1["is_ascending"]==0)&(ds_s1['mask']!=0)).sum(dim='time')
    
    ds_s1_filtered=ds_s1.where(((cnt_ascending>=cnt_descending)&(ds_s1["is_ascending"]==1))|
                               ((cnt_ascending<cnt_descending)&(ds_s1["is_ascending"]==0)))
    # remove intermediate variable
    ds_s1_filtered=ds_s1_filtered.drop_vars(["is_ascending"])
    # drop all-nan time steps
    ds_s1_filtered=ds_s1_filtered.dropna(dim='time',how='all')
    
    return ds_s1_filtered

def lee_filter(da, size):
    """
    Function to apply lee filter of specified window size.
    Adapted from https://stackoverflow.com/questions/39785970/speckle-lee-filter-in-python
    
    Parameters:
    da: Xarray.dataArray, input single-time image data
    size: integer, filtering size in pixels
    
    Returns:
    numpy array of filtered image

    """
    img = da.values
    img_mean = uniform_filter(img, (size, size))
    img_sqr_mean = uniform_filter(img**2, (size, size))
    img_variance = img_sqr_mean - img_mean**2

    overall_variance = variance(img)

    img_weights = img_variance / (img_variance + overall_variance)
    img_output = img_mean + img_weights * (img - img_mean)
    
    return img_output

def preprocess_s1(ds_s1,filter_size=None,s1_orbit_filtering=True):
    '''
    Function to implement preprocessing on Sentinel-1 data, 
    including speckle filtering (optional), filtering observations by orbit (optional) and conversion to dB
    
    Parameters:
    ds_s1: xarray.Dataset
        Time-series of Sentinel-1 data, with variable 'vh' required
    filter_size: integer or None
        Speckle filtering size
    s1_orbit_filtering: Boolean
        Whether to filter Sentinel-1 observations by orbit
        
    Returns:
        xarray.Dataset
        Preprocessed Sentinel-1 data
    '''
    ds_s1_filtered=ds_s1
    
    # apply Lee filtering if required
    if not filter_size is None:
        print('Applying Lee filtering using filtering size of {} pixels...'.format(filter_size))
        # The lee filter above doesn't handle null values
        # We therefore set null values to 0 before applying the filter
        ds_s1_filtered = ds_s1.where(np.isfinite(ds_s1), 0)
        # Create a new entry in dataset corresponding to filtered VV and VH data
        ds_s1_filtered["vh"] = ds_s1_filtered.vh.groupby("time").apply(lee_filter, size=filter_size)
        # Null pixels should remain null, but also including pixels changed to 0 due to the filtering
        ds_s1_filtered['vh'] = ds_s1_filtered.vh.where(ds_s1_filtered.vh!=0,np.nan)
    
    # filter observations by orbit if required
    if s1_orbit_filtering:
        ds_s1_filtered=filter_obs_by_orbit(ds_s1_filtered)

    # Scale to plot data in decibels
    ds_s1_filtered['vh'] = 10 * np.log10(ds_s1_filtered.vh)
    return ds_s1_filtered

def get_mean_number_freq_valid_obs(da,mask,time_step):
    '''
    Calculate mean number of clear observations within each year/timestep in a masked zone
    
    Parameters:
    da: xarray.DataArray
        Time series of a single satellite band/variable (e.g. MNDWI)
    mask: xarray.DataArray or None
        A mask where the calculation is restricted within
    time_step: string
        Pre-defined time step for temporal aggregation, e.g. '1Y'
    
    Returns:
    n_valid_obs: xarray.DataArray
        Average number of valid observations within the coastal zone and for each time step
    freq_valid: xarray.DataArray
        Average frequency of valid observations within the coastal zone and for each time step
    '''
    n_valid_obs=(~da.isnull()).resample(time=time_step).sum('time').compute()
    freq_valid=(~da.isnull()).resample(time=time_step).mean('time').compute()
    if mask is None:
        n_valid_obs=n_valid_obs.mean(dim=['x','y'])
        freq_valid=freq_valid.mean(dim=['x','y'])
    else:
        n_valid_obs=n_valid_obs.where(mask).mean(dim=['x','y'])
        freq_valid=freq_valid.where(mask).mean(dim=['x','y'])
    
    return n_valid_obs,freq_valid

def create_coastal_mask(da,buffer_pixels):
    '''
    Create a simplified coastal zone mask based on time series of Sentinel-2 MNDWI data
    
    Parameters:
    ds_summaries: xarray.DataArray
        Time series of Sentinel-2 MNDWI data
    buffer_pixels: integer
        Number of pixels to buffer coastal zone
    
    Returns:
    coastal_mask: xarray.DataArray 
        A single time buffered coastal zone mask (0: non-coastal and 1: coastal)
    '''
    print('\nCalculating simplified coastal zone mask...')
    # apply thresholding and re-apply nodata values
    nodata = da.isnull()
    thresholded_ds = da>=0
    thresholded_ds = thresholded_ds.where(~nodata)
    # use 20% ~ 80% wet frequency to identify potential coastal zone
    coastal_mask=(thresholded_ds.mean(dim='time') >= 0.2)&(thresholded_ds.mean(dim='time') <= 0.8)
    # buffering
    print('\nApplying buffering of {} Sentinel-2 pixels (parameter buffer_pixels)...'.format(buffer_pixels))
    coastal_mask=xr.apply_ufunc(binary_dilation,coastal_mask.compute(),disk(buffer_pixels))
    return coastal_mask

def choose_product(ds_ls,ds_s2,ds_s1,ds_ls_s2,time_step,**kwargs):
    '''
    Rule-based guide on choosing the best availabel dataset in a given time step and optionally within a coastal zone mask
    
    Parameters:
    ds_ls: xarray.Dataset
        Time series Landsat data
    ds_s2: xarray.Dataset
        Time series Sentinel-2 data
    ds_s1: xarray.Dataset
        Time series Sentinel-1 data
    ds_ls_s2: xarray.Dataset or None
        Time series of combined Landsat and Sentinel-2 data.
    time_step: string
        Time step for temporal composition
    **kwargs: A set of optional parameters including:
        thresh_n_valid: integer
            Threhold of minimum average number of valid observations within each time step
        thresh_freq: float
            Threshold of minimum frequency of valid observations within each time step
        buffer_pixels: integer
            Number of pixels to buffer coastal zone
        coastal_masking: Boolean 
            whether to calculate a coastal zone mask and restrict the comparison of the products within the mask
    Returns:
        Xarray.Dataset of the best product
        String of the best product name: 'ls', 's2', 's1' or 'ls_s2'
    '''
    
    # check if optional parameters are defined otherwise set default values
    thresh_n_valid=10 if "thresh_n_valid" not in kwargs else kwargs["thresh_n_valid"]
    thresh_freq=0.2 if "thresh_freq" not in kwargs else kwargs["thresh_freq"]
    buffer_pixels=100 if "buffer_pixels" not in kwargs else kwargs["buffer_pixels"]
    print('\nThreshold number of valid observations (parameter thresh_n_valid): {}'.format(thresh_n_valid))
    print('\nThreshold frequency of valid observations (parameter thresh_freq): {}'.format(thresh_freq))
    
    # create mask if requested
    coastal_masking=False if "coastal_masking" not in kwargs else kwargs["coastal_masking"]
    if coastal_masking==True:
        # calculate index
        ds_s2 = calculate_indices(ds_s2, index='MNDWI', satellite_mission='s2')
        mask=create_coastal_mask(ds_s2['MNDWI'],buffer_pixels)
    else:
        print('\nNo coastal masking required, using all pixels within the selected region...')
        mask=None

    # calculate mean number and fraction of clear observations within each timestep and the mask
    print('\nCalculating number and frequency of valid observations...')
    
    n_valid_obs_s2,freq_valid_s2=get_mean_number_freq_valid_obs(ds_s2['green'],mask,time_step)
    print('\nSentinel-2: Average number and frequency of valid observations: {:.0f} and {:.2f}'.format(n_valid_obs_s2.mean().values,freq_valid_s2.mean().values))
    
    n_valid_obs_ls,freq_valid_ls=get_mean_number_freq_valid_obs(ds_ls['green'],mask,time_step)
    print('\nLandsat: Average number and frequency of valid observations: {:.0f} and {:.2f}'.format(n_valid_obs_ls.mean().values,freq_valid_ls.mean().values))
    
#     n_valid_obs_s1,freq_valid_s1=get_mean_number_freq_valid_obs(ds_s1['vh'],mask,time_step) # dont need this as sentinel-1 will only be chosen when optical datasets are not sufficient
    if not ds_ls_s2 is None:
        n_valid_obs_ls_s2,freq_valid_ls_s2=get_mean_number_freq_valid_obs(ds_ls_s2['green'],mask,time_step)
        print('\nCombined Landsat and Sentinel-2 product: Average number and frequency of valid observations: {:.0f} and {:.2f}'.format(n_valid_obs_ls_s2.mean().values,freq_valid_ls_s2.mean().values))
        
    # apply decision rules
    print('\nApplying rules to choose product...')
    
    # if Sentinel-2 meets requirements 
    if ((n_valid_obs_s2>=thresh_n_valid).all()) and ((freq_valid_s2>=thresh_freq).all()):
        print('\nSentinel-2 product has met the minimum required average number and frequency of valid observations within all time periods')
        # if combined product is available, choose combined product if it has both higher number and frequency
        if not ds_ls_s2 is None:
            if ((n_valid_obs_ls_s2>n_valid_obs_s2).all()) and ((freq_valid_ls_s2>freq_valid_s2).all()):
                ds_selected, product_name=ds_ls_s2,'ls_s2'
                print('\nChoosing combined Landsat and Sentinel-2 product as it has both higher number and frequency of valid observations within all time periods')
            else:
                ds_selected, product_name=ds_s2,'s2'
                print('\nChoosing Sentinel-2 product as neither Landsat or the combined product meets both requirements or is significantly better than Sentinel-2')
        # if combined product is unavailable, choose Landsat if it has both higher number and frequency
        elif ((n_valid_obs_ls>=n_valid_obs_s2).all()) and ((freq_valid_ls>=freq_valid_s2).all()):
            ds_selected, product_name=ds_ls,'ls'
            print('\nChoosing Landsat product as it has both higher average number and frequency of valid observations within all time periods')
        # otherwise choose Sentinel-2
        else:
            ds_selected, product_name=ds_s2,'s2'
            print('\nChoosing Sentinel-2 product as Landsat product does not meet both requirements or is not significantly better than Sentinel-2')
    # if Sentinel-2 doesn't meet both requirements,but Landsat does, either choose Landsat or combined product if available
    elif ((n_valid_obs_ls>=thresh_n_valid).all()) and ((freq_valid_ls>=thresh_freq).all()):
        print('\nSentinel-2 does not meet the minimum required average number and frequency of valid observations within all time periods, but Landsat does')
        if not ds_ls_s2 is None:
            ds_selected, product_name=ds_ls_s2,'ls_s2'
            print('\nChoosing combined Landsat and Sentinel-2 product as it has both higher number and frequency of valid observations within all time periods')
        else:
            ds_selected, product_name=ds_ls,'ls'
            print('\nChoosing Landsat product')
    # if neither Sentinel-2 or Landsat meet both requirements, choose combined product if it meets requirements
    elif not ds_ls_s2 is None:
        print('\nNeither Sentinel-2 or Landsat meets the minimum required average number and frequency of valid observations within all time periods')
        # but the combined product meet requirements
        if ((n_valid_obs_ls_s2>=thresh_n_valid).all()) and ((freq_valid_ls_s2>=thresh_freq).all()):
            ds_selected, product_name=ds_ls_s2,'ls_s2'
            print('\nChoosing combined Landsat and Sentinel-2 product as it meets the minimum required average number and frequency of valid observations within all time periods')
        else: 
            ds_selected, product_name=ds_s1,'s1'
            print('\nChoosing Sentinel-1 as no other products available that meet the requirements')
    # otherwise choose Sentinel-1
    else:
        print('\nNeither Sentinel-2 or Landsat meets the minimum required average number and frequency of valid observations within all time periods')
        ds_selected, product_name=ds_s1,'s1'
        print('\nChoosing Sentinel-1 product as no other products available that meet the requirements')
        
    print('\nBest available product: ',product_name)
    return ds_selected, product_name

def load_combined_ls_s2(dc,query):
    '''function to query and load combined Landsat and Sentinel-2 data
    
    Parameters:
    dc: connected datacube
    query: a query dictionary to define spatial extent, time range, measurements and spatial resolution for both datasets
    
    Returns:
    ds_combined: Combined data as xarray.Dataset
    '''
    print('Querying and loading combined Landsat and Sentinel-2 products...')
    # Load available Landsat data resampled to Sentinel-2 resolution
    ds_ls = load_ard(dc=dc, products=['ls8_sr', 'ls9_sr'],align=(10, 10),
                     resampling='bilinear',**query)

    # add an variable denoting data source (for future analysis)
    is_ls=xr.DataArray(np.ones(len(ds_ls.time)),dims=('time'),coords={'time': ds_ls.time})
    ds_ls['is_ls'] = is_ls

    # Load Sentinel-2 data
    ds_s2 = load_ard(dc=dc,products=['s2_l2a'],resampling='bilinear',
              align=(10, 10),mask_filters=[("opening", 2), ("dilation", 5)],**query)
    # add an variable denoting data source (for future analysis)
    is_ls=xr.DataArray(np.zeros(len(ds_s2.time)),dims=('time'),coords={'time': ds_s2.time})
    ds_s2['is_ls'] = is_ls

    # merge two datasets together
    ds_combined=xr.concat([ds_ls,ds_s2],dim='time').sortby('time')
    
    return ds_combined

def load_best_available_ds(dc, lat_range, lon_range, time_range, time_step, **kwargs):
    '''
    Function to query, load and compare different products, select and return the best available product
    
    Parameters:
    dc: connected datacube
    lat_range: range of latitudes in tuple or list
    lon_range: range of longitude in tuple or list
    time_range: range of time to query the data in tuple or list
    time_step: string, pre-defined time step for temporal aggregation, e.g. '1Y'
    **kwargs: A set of optional parameters on data query or comparison between products which may include:
    combine_ls_s2: A boolean value indicating whether to include merged/stacked Landsat and Sentinel-2 products as an option. Default to False.
    set_resolution: integer of spatial resolution in metres to query all products
    coastal_masking: A boolean value indicating whether to calculate a mask 
        and restrict the comparison of the products within the masked zone.
    set_product: Set this to only query and load a pre-selected product, 'ls','s2','ls_s2' or 's1'
        i.e. no other products will be queried or compared.
    thresh_n_valid: Threhold of minimum average number of valid observations within each time step, integer
    thresh_freq: Threshold of minimum frequency of valid observations within each time step, float between 0~1
    buffer_pixels: Number of pixels to buffer coastal zone, integer
        
    Returns:
    ds_selected: selected product as xarray.Dataset
    product_name: name of selected product in string format, i.e. 'ls','s2','ls_s2','s1'
    '''
    # parse input time range to accommodate queries before and after 2017
    min_time=min(parser.parse(time_range_i,default=datetime(1987,1,1,0,0)) 
                 for time_range_i in time_range)
    if min_time<datetime(2017,1,1,0,0):
        ls_only=True
    else:
        ls_only=False
        
    # set resolution for query based on optional user input
    if "set_resolution" not in kwargs:
        print('No resolution pre-set, using default resolutions for individual products...')
        resolution_ls=(-30,30)
        resolution_s2=(-10,10)
        resolution_s1=(-20,20)
    else:
        print('Using pre-set spatial resolution for all products...')
        resolution_ls=resolution_s2=resolution_s1=(kwargs["set_resolution"]*(-1),kwargs["set_resolution"])
        
    # create base query for all products
    query = {'x': lon_range,'y': lat_range,'time': time_range,
             'measurements': ['red', 'green', 'blue', 'swir_1'],
             'resolution': resolution_ls, 'group_by':'solar_day',
             'dask_chunks': {'time': 1}}
    
    # Identify the most common projection system in the input query 
    output_crs = mostcommon_crs(dc=dc, product='ls8_sr', query=query)
    
    # update base query
    query.update({'output_crs':output_crs,'min_gooddata':0.2})

    # check if product is pre-set by user
    set_product=None if not "set_product" in kwargs else kwargs["set_product"]
    if not set_product is None:
        product_name=set_product
        
    # check if allowing combining Landsat and Sentinel-2 as an option
    combine_ls_s2=False if not "combine_ls_s2" in kwargs else kwargs["combine_ls_s2"]
    
    # query and load specified products as user provided as possible
    if set_product=='ls':
        print('\nPre-selected product: Landsat')
        ds_selected=load_ard(dc=dc, products=['ls8_sr', 'ls9_sr'],
                             resampling='bilinear',**query)
    elif set_product=='s2':
        print('\nPre-selected product: Sentinel-2')
        if ls_only:
            raise ValueError("Querying date earlier than 2018, please change your pre-selected product as Landsat or query time range.")
        query.update({'resolution': resolution_s2})
        ds_selected= load_ard(dc=dc,products=['s2_l2a'],resampling='bilinear',
                              mask_filters=[("opening", 2), ("dilation", 5)],**query)
    elif set_product=='s1':
        print('\nPre-selected product: Sentinel-1')
        if ls_only:
            raise ValueError("Querying date earlier than 2018, please change your pre-selected product as Landsat or query time range.")
        query.update({'resolution': resolution_s1,'measurements': ['vh','mask']})
        ds_selected=load_s1_by_orbits(dc,query)
    elif set_product=='ls_s2':
        print('\nPre-selected product: combined Landsat and Sentinel-2 products')
        if ("combine_ls_s2" in kwargs)and(combine_ls_s2==False):
            raise ValueError("Conflicting: requesting querying combination of Landsat and Sentinel-2 products while parameter combine_ls_s2 is disabled. Please change parameter and try to run the function again.")
        else:
            query.update({'resolution': resolution_s2})
            ds_selected=load_combined_ls_s2(dc,query)
    else: # no preselection of product or wrong input of product name
        print('\nNo pre-selection of products, querying and compare all products...')
        if ls_only:
            print('\nQuerying date earlier than 2018, only Landsat data will be queried and loaded.')
            # Load available Landsat data
            ds_ls = load_ard(dc=dc, products=['ls8_sr', 'ls9_sr'],
                             resampling='bilinear',**query)
            ds_selected=ds_ls
            product_name='ls'
        else:
            if combine_ls_s2==True:
                query.update({'resolution': resolution_s2})
                ds_ls_s2=load_combined_ls_s2(dc,query)
            else:
                print('\nCombined Landsat and Sentinel-2 products excluded from comparison')
                ds_ls_s2=None
                
            # Load available Landsat data
            print('\nQuerying and loading Landsat data...')
            ds_ls = load_ard(dc=dc, products=['ls8_sr', 'ls9_sr'],
                             resampling='bilinear',**query)
            
            # Load Sentinel-2 data
            print('\nQuerying and Sentinel-2 data...')
            query.update({'resolution': resolution_s2})
            ds_s2 = load_ard(dc=dc,products=['s2_l2a'],resampling='bilinear',
                      mask_filters=[("opening", 2), ("dilation", 5)],**query)
    
            # query and filter Sentinel-1 data by orbit
            query.update({'resolution': resolution_s1,'measurements': ['vh','mask']})
            ds_s1=load_s1_by_orbits(dc,query)
            # apply rules to choose best product
            ds_selected,product_name=choose_product(ds_ls,ds_s2,ds_s1,ds_ls_s2,time_step,**kwargs)
    
    # drop all-nan time steps
    ds_selected=ds_selected.dropna(dim='time',how='all')
    
    return ds_selected,product_name


"""
Functions for calculating per-pixel temporal summary statistics on a
timeseries stored in a xarray.DataArray.

The key functions are:

.. autosummary::
   :caption: Primary functions
   :nosignatures:
   :toctree: gen

   xr_phenology
   temporal_statistics

.. autosummary::
   :nosignatures:
   :toctree: gen

"""

import sys
import dask
import numpy as np
import xarray as xr
import hdstats
from packaging import version
from datacube.utils.geometry import assign_crs


def allNaN_arg(da, dim, stat):
    """
    Calculate da.argmax() or da.argmin() while handling
    all-NaN slices. Fills all-NaN locations with an
    float and then masks the offending cells.

    Parameters
    ----------
    da : xarray.DataArray
    dim : str
        Dimension over which to calculate argmax, argmin e.g. 'time'
    stat : str
        The statistic to calculte, either 'min' for argmin()
        or 'max' for .argmax()

    Returns
    -------
    xarray.DataArray
    """
    # generate a mask where entire axis along dimension is NaN
    mask = da.isnull().all(dim)

    if stat == "max":
        y = da.fillna(float(da.min() - 1))
        y = y.argmax(dim=dim, skipna=True).where(~mask)
        return y

    if stat == "min":
        y = da.fillna(float(da.max() + 1))
        y = y.argmin(dim=dim, skipna=True).where(~mask)
        return y


def _vpos(da):
    """
    vPOS = Value at peak of season
    """
    return da.max("time")


def _pos(da):
    """
    POS = DOY of peak of season
    """
    return da.isel(time=da.argmax("time")).time.dt.dayofyear


def _trough(da):
    """
    Trough = Minimum value
    """
    return da.min("time")


def _aos(vpos, trough):
    """
    AOS = Amplitude of season
    """
    return vpos - trough


def _vsos(da, pos, method_sos="first"):
    """
    vSOS = Value at the start of season
    Params
    -----
    da : xarray.DataArray
    method_sos : str,
        If 'first' then vSOS is estimated
        as the first positive slope on the
        greening side of the curve. If 'median',
        then vSOS is estimated as the median value
        of the postive slopes on the greening side
        of the curve.
    """
    # select timesteps before peak of season (AKA greening)
    greenup = da.where(da.time < pos.time)
    # find the first order slopes
    green_deriv = greenup.differentiate("time")
    # find where the first order slope is postive
    pos_green_deriv = green_deriv.where(green_deriv > 0)
    # positive slopes on greening side
    pos_greenup = greenup.where(~np.isnan(pos_green_deriv))
    # find the median
    median = pos_greenup.median("time")
    # distance of values from median
    distance = pos_greenup - median

    if method_sos == "first":
        # find index (argmin) where distance is most negative
        idx = allNaN_arg(distance, "time", "min").astype("int16")

    if method_sos == "median":
        # find index (argmin) where distance is smallest absolute value
        idx = allNaN_arg(np.fabs(distance), "time", "min").astype("int16")

    return pos_greenup.isel(time=idx)


def _sos(vsos):
    """
    SOS = DOY for start of season
    """
    return vsos.time.dt.dayofyear


def _veos(da, pos, method_eos="last"):
    """
    vEOS = Value at the end of season
    Params
    -----
    method_eos : str
        If 'last' then vEOS is estimated
        as the last negative slope on the
        senescing side of the curve. If 'median',
        then vEOS is estimated as the 'median' value
        of the negative slopes on the senescing
        side of the curve.
    """
    # select timesteps before peak of season (AKA greening)
    senesce = da.where(da.time > pos.time)
    # find the first order slopes
    senesce_deriv = senesce.differentiate("time")
    # find where the fst order slope is negative
    neg_senesce_deriv = senesce_deriv.where(~np.isnan(senesce_deriv < 0))
    # negative slopes on senescing side
    neg_senesce = senesce.where(neg_senesce_deriv)
    # find medians
    median = neg_senesce.median("time")
    # distance to the median
    distance = neg_senesce - median

    if method_eos == "last":
        # index where last negative slope occurs
        idx = allNaN_arg(distance, "time", "min").astype("int16")

    if method_eos == "median":
        # index where median occurs
        idx = allNaN_arg(np.fabs(distance), "time", "min").astype("int16")

    return neg_senesce.isel(time=idx)


def _eos(veos):
    """
    EOS = DOY for end of seasonn
    """
    return veos.time.dt.dayofyear


def _los(da, eos, sos):
    """
    LOS = Length of season (in DOY)
    """
    los = eos - sos
    #handle negative values
    los = xr.where(
        los >= 0,
        los,
        da.time.dt.dayofyear.values[-1] + (eos.where(los < 0) - sos.where(los < 0)),
    )

    return los


def _rog(vpos, vsos, pos, sos):
    """
    ROG = Rate of Greening (Days)
    """
    return (vpos - vsos) / (pos - sos)


def _ros(veos, vpos, eos, pos):
    """
    ROG = Rate of Senescing (Days)
    """
    return (veos - vpos) / (eos - pos)


def xr_phenology(
    da,
    stats=[
        "SOS",
        "POS",
        "EOS",
        "Trough",
        "vSOS",
        "vPOS",
        "vEOS",
        "LOS",
        "AOS",
        "ROG",
        "ROS",
    ],
    method_sos="first",
    method_eos="last",
    verbose=True
):
    """
    Obtain land surface phenology metrics from an
    xarray.DataArray containing a timeseries of a
    vegetation index like NDVI.

    last modified June 2020

    Parameters
    ----------
    da :  xarray.DataArray
        DataArray should contain a 2D or 3D time series of a
        vegetation index like NDVI, EVI
    stats : list
        list of phenological statistics to return. Regardless of
        the metrics returned, all statistics are calculated
        due to inter-dependencies between metrics.
        Options include:

        * `SOS` = DOY of start of season
        * `POS` = DOY of peak of season
        * `EOS` = DOY of end of season
        * `vSOS` = Value at start of season
        * `vPOS` = Value at peak of season
        * `vEOS` = Value at end of season
        * `Trough` = Minimum value of season
        * `LOS` = Length of season (DOY)
        * `AOS` = Amplitude of season (in value units)
        * `ROG` = Rate of greening
        * `ROS` = Rate of senescence

    method_sos : str
        If 'first' then vSOS is estimated as the first positive
        slope on the greening side of the curve. If 'median',
        then vSOS is estimated as the median value of the postive
        slopes on the greening side of the curve.
    method_eos : str
        If 'last' then vEOS is estimated as the last negative slope
        on the senescing side of the curve. If 'median', then vEOS is
        estimated as the 'median' value of the negative slopes on the
        senescing side of the curve.

    Returns
    -------
    xarray.Dataset
        Dataset containing variables for the selected
        phenology statistics

    """
    # Check inputs before running calculations
    if dask.is_dask_collection(da):
        if version.parse(xr.__version__) < version.parse("0.16.0"):
            raise TypeError(
                "Dask arrays are not currently supported by this function, "
                + "run da.compute() before passing dataArray."
            )
        stats_dtype = {
            "SOS": np.int16,
            "POS": np.int16,
            "EOS": np.int16,
            "Trough": np.float32,
            "vSOS": np.float32,
            "vPOS": np.float32,
            "vEOS": np.float32,
            "LOS": np.int16,
            "AOS": np.float32,
            "ROG": np.float32,
            "ROS": np.float32,
        }
        da_template = da.isel(time=0).drop("time")
        template = xr.Dataset(
            {
                var_name: da_template.astype(var_dtype)
                for var_name, var_dtype in stats_dtype.items()
                if var_name in stats
            }
        )
        da_all_time = da.chunk({"time": -1})

        lazy_phenology = da_all_time.map_blocks(
            xr_phenology,
            kwargs=dict(
                stats=stats,
                method_sos=method_sos,
                method_eos=method_eos,
            ),
            template=xr.Dataset(template),
        )

        try:
            crs = da.geobox.crs
            lazy_phenology = assign_crs(lazy_phenology, str(crs))
        except:
            pass

        return lazy_phenology

    if method_sos not in ("median", "first"):
        raise ValueError("method_sos should be either 'median' or 'first'")

    if method_eos not in ("median", "last"):
        raise ValueError("method_eos should be either 'median' or 'last'")

    # If stats supplied is not a list, convert to list.
    stats = stats if isinstance(stats, list) else [stats]

    # try to grab the crs info
    try:
        crs = da.geobox.crs
    except:
        pass

    # remove any remaining all-NaN pixels
    mask = da.isnull().all("time")
    da = da.where(~mask, other=0)

    # calculate the statistics
    if verbose:
        print("      Phenology...")
    vpos = _vpos(da)
    pos = _pos(da)
    trough = _trough(da)
    aos = _aos(vpos, trough)
    vsos = _vsos(da, pos, method_sos=method_sos)
    sos = _sos(vsos)
    veos = _veos(da, pos, method_eos=method_eos)
    eos = _eos(veos)
    los = _los(da, eos, sos)
    rog = _rog(vpos, vsos, pos, sos)
    ros = _ros(veos, vpos, eos, pos)

    # Dictionary containing the statistics
    stats_dict = {
        "SOS": sos.astype(np.int16),
        "EOS": eos.astype(np.int16),
        "vSOS": vsos.astype(np.float32),
        "vPOS": vpos.astype(np.float32),
        "Trough": trough.astype(np.float32),
        "POS": pos.astype(np.int16),
        "vEOS": veos.astype(np.float32),
        "LOS": los.astype(np.int16),
        "AOS": aos.astype(np.float32),
        "ROG": rog.astype(np.float32),
        "ROS": ros.astype(np.float32),
    }

    # intialise dataset with first statistic
    ds = stats_dict[stats[0]].to_dataset(name=stats[0])

    # add the other stats to the dataset
    for stat in stats[1:]:
        if verbose:
            print("         " + stat)
        stats_keep = stats_dict.get(stat)
        ds[stat] = stats_dict[stat]

    try:
        ds = assign_crs(ds, str(crs))
    except:
        pass

    return ds.drop("time")


def temporal_statistics(da, stats):
    """
    Calculate various generic summary statistics on any timeseries.

    This function uses the hdstats temporal library:
    https://github.com/daleroberts/hdstats/blob/master/hdstats/ts.pyx

    last modified June 2020

    Parameters
    ----------
    da :  xarray.DataArray
        DataArray should contain a 3D time series.
    stats : list
        list of temporal statistics to calculate.
        Options include:

        * 'discordance' =
        * 'f_std' = std of discrete fourier transform coefficients, returns
            three layers: f_std_n1, f_std_n2, f_std_n3
        * 'f_mean' = mean of discrete fourier transform coefficients, returns
            three layers: f_mean_n1, f_mean_n2, f_mean_n3
        * 'f_median' = median of discrete fourier transform coefficients, returns
            three layers: f_median_n1, f_median_n2, f_median_n3
        * 'mean_change' = mean of discrete difference along time dimension
        * 'median_change' = median of discrete difference along time dimension
        * 'abs_change' = mean of absolute discrete difference along time dimension
        * 'complexity' =
        * 'central_diff' =
        * 'num_peaks' : The number of peaks in the timeseries, defined with a local
            window of size 10.  NOTE: This statistic is very slow

    Returns
    -------
    xarray.Dataset
        Dataset containing variables for the selected
        temporal statistics

    """

    # if dask arrays then map the blocks
    if dask.is_dask_collection(da):
        if version.parse(xr.__version__) < version.parse("0.16.0"):
            raise TypeError(
                "Dask arrays are only supported by this function if using, "
                + "xarray v0.16, run da.compute() before passing dataArray."
            )

        # create a template that matches the final datasets dims & vars
        arr = da.isel(time=0).drop("time")

        # deal with the case where fourier is first in the list
        if stats[0] in ("f_std", "f_median", "f_mean"):
            template = xr.zeros_like(arr).to_dataset(name=stats[0] + "_n1")
            template[stats[0] + "_n2"] = xr.zeros_like(arr)
            template[stats[0] + "_n3"] = xr.zeros_like(arr)

            for stat in stats[1:]:
                if stat in ("f_std", "f_median", "f_mean"):
                    template[stat + "_n1"] = xr.zeros_like(arr)
                    template[stat + "_n2"] = xr.zeros_like(arr)
                    template[stat + "_n3"] = xr.zeros_like(arr)
                else:
                    template[stat] = xr.zeros_like(arr)
        else:
            template = xr.zeros_like(arr).to_dataset(name=stats[0])

            for stat in stats:
                if stat in ("f_std", "f_median", "f_mean"):
                    template[stat + "_n1"] = xr.zeros_like(arr)
                    template[stat + "_n2"] = xr.zeros_like(arr)
                    template[stat + "_n3"] = xr.zeros_like(arr)
                else:
                    template[stat] = xr.zeros_like(arr)
        try:
            template = template.drop("spatial_ref")
        except:
            pass

        # ensure the time chunk is set to -1
        da_all_time = da.chunk({"time": -1})

        # apply function across chunks
        lazy_ds = da_all_time.map_blocks(
            temporal_statistics, kwargs={"stats": stats}, template=template
        )

        try:
            crs = da.geobox.crs
            lazy_ds = assign_crs(lazy_ds, str(crs))
        except:
            pass

        return lazy_ds

    # If stats supplied is not a list, convert to list.
    stats = stats if isinstance(stats, list) else [stats]

    # grab all the attributes of the xarray
    x, y, time, attrs = da.x, da.y, da.time, da.attrs

    # deal with any all-NaN pixels by filling with 0's
    mask = da.isnull().all("time")
    da = da.where(~mask, other=0)

    # ensure dim order is correct for functions
    da = da.transpose("y", "x", "time").values

    stats_dict = {
        "discordance": lambda da: hdstats.discordance(da, n=10),
        "f_std": lambda da: hdstats.fourier_std(da, n=3, step=5),
        "f_mean": lambda da: hdstats.fourier_mean(da, n=3, step=5),
        "f_median": lambda da: hdstats.fourier_median(da, n=3, step=5),
        "mean_change": lambda da: hdstats.mean_change(da),
        "median_change": lambda da: hdstats.median_change(da),
        "abs_change": lambda da: hdstats.mean_abs_change(da),
        "complexity": lambda da: hdstats.complexity(da),
        "central_diff": lambda da: hdstats.mean_central_diff(da),
        "num_peaks": lambda da: hdstats.number_peaks(da, 10),
    }

    print("   Statistics:")
    # if one of the fourier functions is first (or only)
    # stat in the list then we need to deal with this
    if stats[0] in ("f_std", "f_median", "f_mean"):
        print("      " + stats[0])
        stat_func = stats_dict.get(str(stats[0]))
        zz = stat_func(da)
        n1 = zz[:, :, 0]
        n2 = zz[:, :, 1]
        n3 = zz[:, :, 2]

        # intialise dataset with first statistic
        ds = xr.DataArray(
            n1, attrs=attrs, coords={"x": x, "y": y}, dims=["y", "x"]
        ).to_dataset(name=stats[0] + "_n1")

        # add other datasets
        for i, j in zip([n2, n3], ["n2", "n3"]):
            ds[stats[0] + "_" + j] = xr.DataArray(
                i, attrs=attrs, coords={"x": x, "y": y}, dims=["y", "x"]
            )
    else:
        # simpler if first function isn't fourier transform
        first_func = stats_dict.get(str(stats[0]))
        print("      " + stats[0])
        ds = first_func(da)

        # convert back to xarray dataset
        ds = xr.DataArray(
            ds, attrs=attrs, coords={"x": x, "y": y}, dims=["y", "x"]
        ).to_dataset(name=stats[0])

    # loop through the other functions
    for stat in stats[1:]:
        print("      " + stat)

        # handle the fourier transform examples
        if stat in ("f_std", "f_median", "f_mean"):
            stat_func = stats_dict.get(str(stat))
            zz = stat_func(da)
            n1 = zz[:, :, 0]
            n2 = zz[:, :, 1]
            n3 = zz[:, :, 2]

            for i, j in zip([n1, n2, n3], ["n1", "n2", "n3"]):
                ds[stat + "_" + j] = xr.DataArray(
                    i, attrs=attrs, coords={"x": x, "y": y}, dims=["y", "x"]
                )

        else:
            # Select a stats function from the dictionary
            # and add to the dataset
            stat_func = stats_dict.get(str(stat))
            ds[stat] = xr.DataArray(
                stat_func(da), attrs=attrs, coords={"x": x, "y": y}, dims=["y", "x"]
            )

    # try to add back the geobox
    try:
        crs = da.geobox.crs
        ds = assign_crs(ds, str(crs))
    except:
        pass

    return ds
